@@ -167,7 +167,7 @@ namespace ikvm.awt
             // this is required to simulate Graphics.create(), which is not possible in .NET
             // we simply call Save on create() an restore this state, if any method is called
             // on the current graphics. This will work for almost any use case of create()
-            if (baseContext.Current != this)
+            if (baseContext != null && baseContext.Current != this)
             {
                 if (!baseContext.Current.disposed)
                 {
@@ -750,6 +750,7 @@ namespace ikvm.awt
             {
                 netG.pen = new Pen(netG.color);
                 netG.brush = new SolidBrush(netG.color);
+                netG.setRenderingHint(java.awt.RenderingHints.KEY_TEXT_ANTIALIASING, java.awt.RenderingHints.VALUE_TEXT_ANTIALIAS_DEFAULT);
             }
         }
     }
@@ -770,6 +771,8 @@ namespace ikvm.awt
         internal Pen pen;
         private CompositeHelper composite;
         private java.awt.Composite javaComposite = java.awt.AlphaComposite.SrcOver;
+        private Object textAntialiasHint;
+        private Object fractionalHint = java.awt.RenderingHints.VALUE_FRACTIONALMETRICS_DEFAULT;
 
         protected NetGraphics(Graphics g, java.awt.Font font, Color fgcolor, Color bgcolor)
         {
@@ -1265,28 +1268,32 @@ namespace ikvm.awt
             drawString(str, (float)x, (float)y);
         }
 
-        public override void drawString(string text, float x, float y)
-		{
-            if (text.Length == 0) {
+        public override void drawString(String text, float x, float y)
+        {
+            if (text.Length == 0)
+            {
                 return;
             }
-            StringFormat format;
-            switch (g.TextRenderingHint)
-            {
-                // Fractional metrics
-                case TextRenderingHint.AntiAlias:
-                case TextRenderingHint.SingleBitPerPixel:
-                    // this very mystic, if a StringFormat extends from GenericTypographic then the metric are different but like Java with fractional metrics
-                    format = new StringFormat(StringFormat.GenericTypographic);
-                    break;
-                default:
-                    // Non Fractional metrics
-                    format = new StringFormat();
-                    break;
-            }
+            bool fractional = isFractionalMetrics();
+            StringFormat format = new StringFormat(StringFormat.GenericTypographic);
             format.FormatFlags = StringFormatFlags.MeasureTrailingSpaces | StringFormatFlags.NoWrap | StringFormatFlags.FitBlackBox;
             format.Trimming = StringTrimming.None;
-            g.DrawString(text, netfont, brush, x, y - font.getSize(), format);
+            if (fractional || !sun.font.StandardGlyphVector.isSimpleString(font, text))
+            {
+                g.DrawString(text, netfont, brush, x, y - font.getSize(), format);
+            }
+            else
+            {
+                // fixed metric for simple text, we position every character to simulate the Java behaviour
+                java.awt.font.FontRenderContext frc = new java.awt.font.FontRenderContext(null, isAntiAlias(), fractional);
+                sun.font.FontDesignMetrics metrics = sun.font.FontDesignMetrics.getMetrics(font, frc);
+                y -= font.getSize();
+                for (int i = 0; i < text.Length; i++)
+                {
+                    g.DrawString(text.Substring(i, 1), netfont, brush, x, y, format);
+                    x += metrics.charWidth(text[i]);
+                }
+            }
         }
 
         public override void drawString(java.text.AttributedCharacterIterator iterator, int x, int y)
@@ -1630,62 +1637,30 @@ namespace ikvm.awt
             }
             if (hintKey == java.awt.RenderingHints.KEY_TEXT_ANTIALIASING)
             {
-                if (hintValue == java.awt.RenderingHints.VALUE_TEXT_ANTIALIAS_DEFAULT || hintValue == java.awt.RenderingHints.VALUE_TEXT_ANTIALIAS_OFF)
+                if (hintValue == java.awt.RenderingHints.VALUE_TEXT_ANTIALIAS_DEFAULT ||
+                    hintValue == java.awt.RenderingHints.VALUE_TEXT_ANTIALIAS_OFF)
                 {
-                    switch (g.TextRenderingHint) {
-                        case TextRenderingHint.SystemDefault:
-                        case TextRenderingHint.AntiAlias:
-                            g.TextRenderingHint = TextRenderingHint.SingleBitPerPixel;
-                            break;
-                        case TextRenderingHint.AntiAliasGridFit:
-                            g.TextRenderingHint = TextRenderingHint.SingleBitPerPixelGridFit;
-                            break;
-                    }
+                    g.TextRenderingHint = TextRenderingHint.SingleBitPerPixelGridFit;
+                    textAntialiasHint = hintValue;
                     return;
                 }
                 if (hintValue == java.awt.RenderingHints.VALUE_TEXT_ANTIALIAS_ON)
                 {
-                    switch (g.TextRenderingHint) {
-                        case TextRenderingHint.SystemDefault:
-                        case TextRenderingHint.SingleBitPerPixel:
-                            g.TextRenderingHint = TextRenderingHint.AntiAlias;
-                            break;
-                        case TextRenderingHint.SingleBitPerPixelGridFit:
-                            g.TextRenderingHint = TextRenderingHint.AntiAliasGridFit;
-                            break;
-                    }
+                    g.TextRenderingHint = TextRenderingHint.AntiAlias;
+                    textAntialiasHint = hintValue;
                     return;
                 }
                 return;
             }
             if (hintKey == java.awt.RenderingHints.KEY_FRACTIONALMETRICS) 
             {
-                if (hintValue == java.awt.RenderingHints.VALUE_FRACTIONALMETRICS_DEFAULT || hintValue == java.awt.RenderingHints.VALUE_FRACTIONALMETRICS_OFF) 
+                if (hintValue == java.awt.RenderingHints.VALUE_FRACTIONALMETRICS_DEFAULT || 
+                    hintValue == java.awt.RenderingHints.VALUE_FRACTIONALMETRICS_OFF ||
+                    hintValue == java.awt.RenderingHints.VALUE_FRACTIONALMETRICS_ON) 
                 {
-                    // OFF means enable GridFit
-                    switch (g.TextRenderingHint) {
-                        case TextRenderingHint.SystemDefault:
-                        case TextRenderingHint.SingleBitPerPixel:
-                            g.TextRenderingHint = TextRenderingHint.SingleBitPerPixelGridFit;
-                            break;
-                        case TextRenderingHint.AntiAlias:
-                            g.TextRenderingHint = TextRenderingHint.AntiAliasGridFit;
-                            break;
-                    }
-                }
-                if (hintValue == java.awt.RenderingHints.VALUE_FRACTIONALMETRICS_ON) 
-                {
-                    // ON means remove GridFit
-                    switch (g.TextRenderingHint) {
-                        case TextRenderingHint.SingleBitPerPixelGridFit:
-                            g.TextRenderingHint = TextRenderingHint.SingleBitPerPixel;
-                            break;
-                        case TextRenderingHint.AntiAliasGridFit:
-                        case TextRenderingHint.ClearTypeGridFit:
-                            g.TextRenderingHint = TextRenderingHint.AntiAlias;
-                            break;
-                    }
+                    fractionalHint = hintValue;
                 }
+                return;
             }
 
         }
@@ -1742,30 +1717,8 @@ namespace ikvm.awt
                     break;
             }
 
-            switch (g.TextRenderingHint)
-            {
-                case TextRenderingHint.SystemDefault:
-                    hints.put(java.awt.RenderingHints.KEY_TEXT_ANTIALIASING, java.awt.RenderingHints.VALUE_TEXT_ANTIALIAS_DEFAULT);
-                    hints.put(java.awt.RenderingHints.KEY_FRACTIONALMETRICS, java.awt.RenderingHints.VALUE_FRACTIONALMETRICS_DEFAULT);
-                    break;
-                case TextRenderingHint.SingleBitPerPixelGridFit:
-                    hints.put(java.awt.RenderingHints.KEY_TEXT_ANTIALIASING, java.awt.RenderingHints.VALUE_TEXT_ANTIALIAS_OFF);
-                    hints.put(java.awt.RenderingHints.KEY_FRACTIONALMETRICS, java.awt.RenderingHints.VALUE_FRACTIONALMETRICS_OFF);
-                    break;
-                case TextRenderingHint.SingleBitPerPixel:
-                    hints.put(java.awt.RenderingHints.KEY_TEXT_ANTIALIASING, java.awt.RenderingHints.VALUE_TEXT_ANTIALIAS_OFF);
-                    hints.put(java.awt.RenderingHints.KEY_FRACTIONALMETRICS, java.awt.RenderingHints.VALUE_FRACTIONALMETRICS_ON);
-                    break;
-                case TextRenderingHint.AntiAlias:
-                    hints.put(java.awt.RenderingHints.KEY_TEXT_ANTIALIASING, java.awt.RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
-                    hints.put(java.awt.RenderingHints.KEY_FRACTIONALMETRICS, java.awt.RenderingHints.VALUE_FRACTIONALMETRICS_ON);
-                    break;
-                case TextRenderingHint.AntiAliasGridFit:
-                case TextRenderingHint.ClearTypeGridFit:
-                    hints.put(java.awt.RenderingHints.KEY_TEXT_ANTIALIASING, java.awt.RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
-                    hints.put(java.awt.RenderingHints.KEY_FRACTIONALMETRICS, java.awt.RenderingHints.VALUE_FRACTIONALMETRICS_OFF);
-                    break;
-            }
+            hints.put(java.awt.RenderingHints.KEY_TEXT_ANTIALIASING, textAntialiasHint);
+            hints.put(java.awt.RenderingHints.KEY_FRACTIONALMETRICS, fractionalHint);
             return hints;
         }
 
@@ -1870,9 +1823,27 @@ namespace ikvm.awt
             return stroke; 
         }
 
+        private bool isAntiAlias()
+        {
+            switch (g.TextRenderingHint)
+            {
+                case TextRenderingHint.AntiAlias:
+                case TextRenderingHint.AntiAliasGridFit:
+                case TextRenderingHint.ClearTypeGridFit:
+                    return true;
+                default:
+                    return false;
+            }
+        }
+
+        private bool isFractionalMetrics()
+        {
+            return fractionalHint == java.awt.RenderingHints.VALUE_FRACTIONALMETRICS_ON; 
+        }
+
         public override java.awt.font.FontRenderContext getFontRenderContext()
         {
-            return new java.awt.font.FontRenderContext(getTransform(), false, false);
+            return new java.awt.font.FontRenderContext(getTransform(), isAntiAlias(), isFractionalMetrics());
         }
 
         public override void drawGlyphVector(java.awt.font.GlyphVector gv, float x, float y)
