@@ -304,8 +304,13 @@ namespace IKVM.Internal
 			}
 		}
 
-		protected override bool EmitMapXmlMethodBody(CodeEmitter ilgen, ClassFile f, ClassFile.Method m)
+		protected override bool EmitMapXmlMethodPrologueAndOrBody(CodeEmitter ilgen, ClassFile f, ClassFile.Method m)
 		{
+			IKVM.Internal.MapXml.InstructionList prologue = classLoader.GetMethodPrologue(new MethodKey(f.Name, m.Name, m.Signature));
+			if(prologue != null)
+			{
+				prologue.Emit(classLoader, ilgen);
+			}
 			Dictionary<MethodKey, IKVM.Internal.MapXml.InstructionList> mapxml = classLoader.GetMapXmlMethodBodies();
 			if(mapxml != null)
 			{
@@ -662,8 +667,9 @@ namespace IKVM.Internal
 								{
 									AttributeHelper.SetModifiers(cb, (Modifiers)constructor.Modifiers, false);
 								}
+								CompilerClassLoader.AddDeclaredExceptions(cb, constructor.throws);
 								CodeEmitter ilgen = CodeEmitter.Create(cb);
-								constructor.body.Emit(classLoader, ilgen);
+								constructor.Emit(classLoader, ilgen);
 								ilgen.DoEmit();
 								if(constructor.Attributes != null)
 								{
@@ -725,8 +731,9 @@ namespace IKVM.Internal
 									mw.Link();
 									typeBuilder.DefineMethodOverride(mb, (MethodInfo)mw.GetMethod());
 								}
+								CompilerClassLoader.AddDeclaredExceptions(mb, method.throws);
 								CodeEmitter ilgen = CodeEmitter.Create(mb);
-								method.body.Emit(classLoader, ilgen);
+								method.Emit(classLoader, ilgen);
 								ilgen.DoEmit();
 								if(method.Attributes != null)
 								{
