@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2002-2010 Jeroen Frijters
+  Copyright (C) 2002-2011 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -289,7 +289,8 @@ class IkvmcCompiler
 		Console.Error.WriteLine("    -target:library            Build a library");
 		Console.Error.WriteLine("    -target:module             Build a module for use by the linker");
 		Console.Error.WriteLine("    -platform:<string>         Limit which platforms this code can run on: x86,");
-		Console.Error.WriteLine("                               Itanium, x64, or anycpu. The default is anycpu.");
+		Console.Error.WriteLine("                               Itanium, x64, arm, anycpu32bitpreferred, or");
+		Console.Error.WriteLine("                               anycpu. The default is anycpu.");
 		Console.Error.WriteLine("    -keyfile:<keyfilename>     Use keyfile to sign the assembly");
 		Console.Error.WriteLine("    -key:<keycontainer>        Use keycontainer to sign the assembly");
 		Console.Error.WriteLine("    -delaysign                 Delay-sign the assembly");
@@ -330,8 +331,8 @@ class IkvmcCompiler
 		Console.Error.WriteLine("    -privatepackage:<prefix>   Mark all classes with a package name starting");
 		Console.Error.WriteLine("                               with <prefix> as internal to the assembly");
 		Console.Error.WriteLine("    -nowarn:<warning[:key]>    Suppress specified warnings");
-		Console.Error.WriteLine("    -warnaserror:<warning[:key]>");
-		Console.Error.WriteLine("                               Treat specified warnings as errors");
+		Console.Error.WriteLine("    -warnaserror[:<warning[:key]>]");
+		Console.Error.WriteLine("                               Treat (specified) warnings as errors");
 		Console.Error.WriteLine("    -writeSuppressWarningsFile:<file>");
 		Console.Error.WriteLine("                               Write response file with -nowarn:<warning[:key]>");
 		Console.Error.WriteLine("                               options to suppress all encountered warnings");
@@ -445,6 +446,14 @@ class IkvmcCompiler
 							options.pekind = PortableExecutableKinds.ILOnly | PortableExecutableKinds.PE32Plus;
 							options.imageFileMachine = ImageFileMachine.AMD64;
 							break;
+						case "-platform:arm":
+							options.pekind = PortableExecutableKinds.ILOnly;
+							options.imageFileMachine = ImageFileMachine.ARM;
+							break;
+						case "-platform:anycpu32bitpreferred":
+							options.pekind = PortableExecutableKinds.ILOnly | PortableExecutableKinds.Preferred32Bit;
+							options.imageFileMachine = ImageFileMachine.I386;
+							break;
 						case "-platform:anycpu":
 							options.pekind = PortableExecutableKinds.ILOnly;
 							options.imageFileMachine = ImageFileMachine.I386;
@@ -699,6 +708,10 @@ class IkvmcCompiler
 						options.suppressWarnings[ws] = ws;
 					}
 				}
+				else if(s == "-warnaserror")
+				{
+					options.warnaserror = true;
+				}
 				else if(s.StartsWith("-warnaserror:"))
 				{
 					foreach(string w in s.Substring(13).Split(','))
@@ -778,6 +791,15 @@ class IkvmcCompiler
 						return 1;
 					}
 				}
+				else if(s.StartsWith("-proxy:")) // currently undocumented!
+				{
+					string proxy = s.Substring(7);
+					if(options.proxies.Contains(proxy))
+					{
+						StaticCompiler.IssueMessage(Message.DuplicateProxy, proxy);
+					}
+					options.proxies.Add(proxy);
+				}
 				else
 				{
 					Console.Error.WriteLine("Error: unrecognized option: {0}", s);
