@@ -71,6 +71,38 @@ namespace IKVM.Internal.MapXml
 		}
 
 		internal abstract void Generate(CodeGenContext context, CodeEmitter ilgen);
+
+		public override string ToString()
+		{
+			System.Text.StringBuilder sb = new System.Text.StringBuilder();
+			sb.Append('<');
+			object[] attr = GetType().GetCustomAttributes(typeof(XmlTypeAttribute), false);
+			if (attr.Length == 1)
+			{
+				sb.Append(((XmlTypeAttribute)attr[0]).TypeName);
+			}
+			else
+			{
+				sb.Append(GetType().Name);
+			}
+			foreach (System.Reflection.FieldInfo field in GetType().GetFields())
+			{
+				if (!field.IsStatic)
+				{
+					object value = field.GetValue(this);
+					if (value != null)
+					{
+						attr = field.GetCustomAttributes(typeof(XmlAttributeAttribute), false);
+						if (attr.Length == 1)
+						{
+							sb.AppendFormat(" {0}=\"{1}\"", ((XmlAttributeAttribute)attr[0]).AttributeName, value);
+						}
+					}
+				}
+			}
+			sb.Append(" />");
+			return sb.ToString();
+		}
 	}
 
 	[XmlType("ldstr")]
@@ -630,6 +662,18 @@ namespace IKVM.Internal.MapXml
 		}
 	}
 
+	[XmlType("ldarg_s")]
+	public sealed class LdArg_S : Instruction
+	{
+		[XmlAttribute("argNum")]
+		public byte ArgNum;
+
+		internal override void Generate(CodeGenContext context, CodeEmitter ilgen)
+		{
+			ilgen.Emit(OpCodes.Ldarg_S, ArgNum);
+		}
+	}
+
 	[XmlType("ldarg_0")]
 	public sealed class LdArg_0 : Simple
 	{
@@ -742,6 +786,15 @@ namespace IKVM.Internal.MapXml
 		}
 	}
 
+	[XmlType("stind_i8")]
+	public sealed class Stind_i8 : Simple
+	{
+		public Stind_i8()
+			: base(OpCodes.Stind_I8)
+		{
+		}
+	}
+
 	[XmlType("stind_ref")]
 	public sealed class Stind_ref : Simple
 	{
@@ -1019,6 +1072,33 @@ namespace IKVM.Internal.MapXml
 		}
 	}
 
+	[XmlType("or")]
+	public sealed class Or : Simple
+	{
+		public Or()
+			: base(OpCodes.Or)
+		{
+		}
+	}
+
+	[XmlType("xor")]
+	public sealed class Xor : Simple
+	{
+		public Xor()
+			: base(OpCodes.Xor)
+		{
+		}
+	}
+
+	[XmlType("not")]
+	public sealed class Not : Simple
+	{
+		public Not()
+			: base(OpCodes.Not)
+		{
+		}
+	}
+
 	[XmlType("unaligned")]
 	public sealed class Unaligned : Instruction
 	{
@@ -1158,10 +1238,129 @@ namespace IKVM.Internal.MapXml
 	{
 		[XmlAttribute("type")]
 		public string type;
+		[XmlAttribute("class")]
+		public string Class;
+		[XmlAttribute("method")]
+		public string Method;
+		[XmlAttribute("field")]
+		public string Field;
+		[XmlAttribute("sig")]
+		public string Sig;
 
 		internal override void Generate(CodeGenContext context, CodeEmitter ilgen)
 		{
-			ilgen.Emit(OpCodes.Ldtoken, StaticCompiler.GetTypeForMapXml(context.ClassLoader, type));
+			if (!Validate())
+			{
+				return;
+			}
+
+			MemberInfo member = Resolve(context);
+			Type type = member as Type;
+			MethodInfo method = member as MethodInfo;
+			ConstructorInfo constructor = member as ConstructorInfo;
+			FieldInfo field = member as FieldInfo;
+
+			if (type != null)
+			{
+				ilgen.Emit(OpCodes.Ldtoken, type);
+			}
+			else if (method != null)
+			{
+				ilgen.Emit(OpCodes.Ldtoken, method);
+			}
+			else if (constructor != null)
+			{
+				ilgen.Emit(OpCodes.Ldtoken, constructor);
+			}
+			else if (field != null)
+			{
+				ilgen.Emit(OpCodes.Ldtoken, field);
+			}
+			else
+			{
+				StaticCompiler.IssueMessage(Message.MapXmlUnableToResolveOpCode, ToString());
+			}
+		}
+
+		private bool Validate()
+		{
+			if (type != null && Class == null)
+			{
+				if (Method != null || Field != null || Sig != null)
+				{
+					StaticCompiler.IssueMessage(Message.MapXmlError, "not implemented: cannot use 'type' attribute with 'method' or 'field' attribute for ldtoken");
+					return false;
+				}
+				return true;
+			}
+			else if (Class != null && type == null)
+			{
+				if (Method == null && Field == null)
+				{
+					if (Sig != null)
+					{
+						StaticCompiler.IssueMessage(Message.MapXmlError, "cannot specify 'sig' attribute without either 'method' or 'field' attribute for ldtoken");
+					}
+					return true;
+				}
+				if (Method != null && Field != null)
+				{
+					StaticCompiler.IssueMessage(Message.MapXmlError, "cannot specify both 'method' and 'field' attribute for ldtoken");
+					return false;
+				}
+				return true;
+			}
+			else
+			{
+				StaticCompiler.IssueMessage(Message.MapXmlError, "must specify either 'type' or 'class' attribute for ldtoken");
+				return false;
+			}
+		}
+
+		private MemberInfo Resolve(CodeGenContext context)
+		{
+			if (type != null)
+			{
+				if (Class != null || Method != null || Field != null || Sig != null)
+				{
+					throw new NotImplementedException();
+				}
+				return StaticCompiler.GetTypeForMapXml(context.ClassLoader, type);
+			}
+			else if (Class != null)
+			{
+				TypeWrapper tw = context.ClassLoader.LoadClassByDottedNameFast(Class);
+				if (tw == null)
+				{
+					return null;
+				}
+				else if (Method != null)
+				{
+					MethodWrapper mw = tw.GetMethodWrapper(Method, Sig, false);
+					if (mw == null)
+					{
+						return null;
+					}
+					return mw.GetMethod();
+				}
+				else if (Field != null)
+				{
+					FieldWrapper fw = tw.GetFieldWrapper(Field, Sig);
+					if (fw == null)
+					{
+						return null;
+					}
+					return fw.GetField();
+				}
+				else
+				{
+					return tw.TypeAsBaseType;
+				}
+			}
+			else
+			{
+				return null;
+			}
 		}
 	}
 
@@ -1218,6 +1417,7 @@ namespace IKVM.Internal.MapXml
 		[XmlElement(typeof(StLoc))]
 		[XmlElement(typeof(LdLoc))]
 		[XmlElement(typeof(LdArga))]
+		[XmlElement(typeof(LdArg_S))]
 		[XmlElement(typeof(LdArg_0))]
 		[XmlElement(typeof(LdArg_1))]
 		[XmlElement(typeof(LdArg_2))]
@@ -1232,6 +1432,7 @@ namespace IKVM.Internal.MapXml
 		[XmlElement(typeof(Stind_i1))]
 		[XmlElement(typeof(Stind_i2))]
 		[XmlElement(typeof(Stind_i4))]
+		[XmlElement(typeof(Stind_i8))]
 		[XmlElement(typeof(Stind_ref))]
 		[XmlElement(typeof(Ret))]
 		[XmlElement(typeof(Throw))]
@@ -1260,6 +1461,9 @@ namespace IKVM.Internal.MapXml
 		[XmlElement(typeof(Sub))]
 		[XmlElement(typeof(Mul))]
 		[XmlElement(typeof(And))]
+		[XmlElement(typeof(Or))]
+		[XmlElement(typeof(Xor))]
+		[XmlElement(typeof(Not))]
 		[XmlElement(typeof(Unaligned))]
 		[XmlElement(typeof(Cpblk))]
 		[XmlElement(typeof(Ceq))]
@@ -1301,23 +1505,6 @@ namespace IKVM.Internal.MapXml
 		public string Class;
 	}
 
-	public class Constructor
-	{
-		[XmlAttribute("sig")]
-		public string Sig;
-		[XmlAttribute("modifiers")]
-		public MapModifiers Modifiers;
-		[XmlElement("parameter")]
-		public Param[] Params;
-		public InstructionList body;
-		public InstructionList alternateBody;
-		public Redirect redirect;
-		[XmlElement("throws", typeof(Throws))]
-		public Throws[] throws;
-		[XmlElement("attribute")]
-		public Attribute[] Attributes;
-	}
-
 	public class Redirect
 	{
 		private int linenum = Root.LineNumber;
@@ -1396,34 +1583,39 @@ namespace IKVM.Internal.MapXml
 		public InstructionList code;
 	}
 
-	public class Method
+	public abstract class MethodBase
 	{
-		[XmlAttribute("name")]
-		public string Name;
-		[XmlAttribute("sig")]
-		public string Sig;
-		[XmlAttribute("modifiers")]
-		public MapModifiers Modifiers;
 		[XmlAttribute("attributes")]
 		public MethodAttributes MethodAttributes;
-		[XmlAttribute("nonullcheck")]
-		public bool NoNullCheck;
-		[XmlElement("parameter")]
-		public Param[] Params;
 		public InstructionList body;
-		public InstructionList alternateBody;
-		public InstructionList nonvirtualAlternateBody;
-		public Redirect redirect;
-		public Override @override;
 		[XmlElement("throws", typeof(Throws))]
 		public Throws[] throws;
 		[XmlElement("attribute")]
 		public Attribute[] Attributes;
 		[XmlElement("replace-method-call")]
 		public ReplaceMethodCall[] ReplaceMethodCalls;
+		public InstructionList prologue;
+
+		internal abstract MethodKey ToMethodKey(string className);
+	}
+
+	public abstract class MethodConstructorBase : MethodBase
+	{
+		[XmlAttribute("sig")]
+		public string Sig;
+		[XmlAttribute("modifiers")]
+		public MapModifiers Modifiers;
+		[XmlElement("parameter")]
+		public Param[] Params;
+		public InstructionList alternateBody;
+		public Redirect redirect;
 
 		internal void Emit(ClassLoaderWrapper loader, CodeEmitter ilgen)
 		{
+			if(prologue != null)
+			{
+				prologue.Emit(loader, ilgen);
+			}
 			if(redirect != null)
 			{
 				redirect.Emit(loader, ilgen);
@@ -1435,6 +1627,37 @@ namespace IKVM.Internal.MapXml
 		}
 	}
 
+	public class Method : MethodConstructorBase
+	{
+		[XmlAttribute("name")]
+		public string Name;
+		[XmlAttribute("nonullcheck")]
+		public bool NoNullCheck;
+		public InstructionList nonvirtualAlternateBody;
+		public Override @override;
+
+		internal override MethodKey ToMethodKey(string className)
+		{
+			return new MethodKey(className, Name, Sig);
+		}
+	}
+
+	public class Constructor : MethodConstructorBase
+	{
+		internal override MethodKey ToMethodKey(string className)
+		{
+			return new MethodKey(className, StringConstants.INIT, Sig);
+		}
+	}
+
+	public class ClassInitializer : MethodBase
+	{
+		internal override MethodKey ToMethodKey(string className)
+		{
+			return new MethodKey(className, StringConstants.CLINIT, StringConstants.SIG_VOID);
+		}
+	}
+
 	public class Field
 	{
 		[XmlAttribute("name")]
@@ -1555,7 +1778,7 @@ namespace IKVM.Internal.MapXml
 		[XmlElement("implements")]
 		public Interface[] Interfaces;
 		[XmlElement("clinit")]
-		public Method Clinit;
+		public ClassInitializer Clinit;
 		[XmlElement("attribute")]
 		public Attribute[] Attributes;
 	}
