@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2002-2010 Jeroen Frijters
+  Copyright (C) 2002-2011 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -52,6 +52,7 @@ static class NetExp
 		bool nostdlib = false;
 		bool bootstrap = false;
 		string outputFile = null;
+		bool forwarders = false;
 		foreach(string s in args)
 		{
 			if(s.StartsWith("-") || assemblyNameOrPath != null)
@@ -92,6 +93,10 @@ static class NetExp
 				{
 					namespaces.Add(s.Substring(11) + ".");
 				}
+				else if(s == "-forwarders")
+				{
+					forwarders = true;
+				}
 				else
 				{
 					// unrecognized option, or multiple assemblies, print usage message and exit
@@ -119,6 +124,7 @@ static class NetExp
 			Console.Error.WriteLine("    -nostdlib                  Do not reference standard libraries");
 			Console.Error.WriteLine("    -lib:<dir>                 Additional directories to search for references");
 			Console.Error.WriteLine("    -namespace:<ns>            Only include types from specified namespace");
+			Console.Error.WriteLine("    -forwarders                Export forwarded types too");
 			return 1;
 		}
 		if(File.Exists(assemblyNameOrPath) && nostdlib)
@@ -198,7 +204,15 @@ static class NetExp
 						}
 						foreach (Assembly asm in assemblies)
 						{
-							if (ProcessAssembly(asm, continueOnError) != 0)
+							if (ProcessTypes(asm.GetTypes(), continueOnError) != 0)
+							{
+								rc = 1;
+								if (!continueOnError)
+								{
+									break;
+								}
+							}
+							if (forwarders && ProcessTypes(asm.ManifestModule.__GetExportedTypes(), continueOnError) != 0)
 							{
 								rc = 1;
 								if (!continueOnError)
@@ -208,13 +222,9 @@ static class NetExp
 							}
 						}
 					}
-					catch (TypeLoadException x)
-					{
-						Console.WriteLine(x);
-					}
 					catch (System.Exception x)
 					{
-						Console.WriteLine(x);
+						Console.Error.WriteLine(x);
 						
 						if (!continueOnError)
 						{
@@ -400,7 +410,20 @@ static class NetExp
 				if (mb != null)
 				{
 					ThrowsAttribute throws = AttributeHelper.GetThrows(mb);
-					if (throws != null)
+					if (throws == null)
+					{
+						string[] throwsArray = mw.GetDeclaredExceptions();
+						if (throwsArray != null && throwsArray.Length > 0)
+						{
+							IKVM.StubGen.ExceptionsAttribute attrib = new IKVM.StubGen.ExceptionsAttribute(writer);
+							foreach (string ex in throwsArray)
+							{
+								attrib.Add(ex.Replace('.', '/'));
+							}
+							m.AddAttribute(attrib);
+						}
+					}
+					else
 					{
 						IKVM.StubGen.ExceptionsAttribute attrib = new IKVM.StubGen.ExceptionsAttribute(writer);
 						if (throws.classes != null)
@@ -414,7 +437,7 @@ static class NetExp
 						{
 							foreach (Type ex in throws.types)
 							{
-								attrib.Add(ex.FullName.Replace('.', '/'));
+								attrib.Add(ClassLoaderWrapper.GetWrapperFromType(ex).Name.Replace('.', '/'));
 							}
 						}
 						m.AddAttribute(attrib);
@@ -423,6 +446,11 @@ static class NetExp
 					{
 						m.AddAttribute(new IKVM.StubGen.DeprecatedAttribute(writer));
 					}
+					IList<CustomAttributeData> attr = CustomAttributeData.__GetCustomAttributes(mb, JVM.LoadType(typeof(AnnotationDefaultAttribute)), false);
+					if (attr.Count == 1)
+					{
+						m.AddAttribute(new IKVM.StubGen.AnnotationDefaultClassFileAttribute(writer, GetAnnotationDefault(writer, attr[0].ConstructorArguments[0])));
+					}
 				}
 				string sig = tw.GetGenericMethodSignature(mw);
 				if (sig != null)
@@ -470,8 +498,12 @@ static class NetExp
 		}
 		AddMetaAnnotations(writer, tw);
 		zipCount++;
-		zipFile.PutNextEntry(new ZipEntry(name + ".class"));
-		writer.Write(zipFile);
+		MemoryStream mem = new MemoryStream();
+		writer.Write(mem);
+		ZipEntry entry = new ZipEntry(name + ".class");
+		entry.Size = mem.Position;
+		zipFile.PutNextEntry(entry);
+		mem.WriteTo(zipFile);
 	}
 
 	private static string GetAssemblyName(TypeWrapper tw)
@@ -625,6 +657,118 @@ static class NetExp
 		return mem.ToArray();
 	}
 
+	private static byte[] GetAnnotationDefault(IKVM.StubGen.ClassFileWriter classFile, CustomAttributeTypedArgument value)
+	{
+		MemoryStream mem = new MemoryStream();
+		IKVM.StubGen.BigEndianStream bes = new IKVM.StubGen.BigEndianStream(mem);
+		try
+		{
+			WriteAnnotationElementValue(classFile, bes, value);
+		}
+		catch (InvalidCastException)
+		{
+			Console.Error.WriteLine("Warning: incorrect annotation default value");
+		}
+		catch (IndexOutOfRangeException)
+		{
+			Console.Error.WriteLine("Warning: incorrect annotation default value");
+		}
+		return mem.ToArray();
+	}
+
+	private static void WriteAnnotationElementValue(IKVM.StubGen.ClassFileWriter classFile, IKVM.StubGen.BigEndianStream bes, CustomAttributeTypedArgument value)
+	{
+		if (value.ArgumentType == Types.Boolean)
+		{
+			bes.WriteByte((byte)'Z');
+			bes.WriteUInt16(classFile.AddInt((bool)value.Value ? 1 : 0));
+		}
+		else if (value.ArgumentType == Types.Byte)
+		{
+			bes.WriteByte((byte)'B');
+			bes.WriteUInt16(classFile.AddInt((byte)value.Value));
+		}
+		else if (value.ArgumentType == Types.Char)
+		{
+			bes.WriteByte((byte)'C');
+			bes.WriteUInt16(classFile.AddInt((char)value.Value));
+		}
+		else if (value.ArgumentType == Types.Int16)
+		{
+			bes.WriteByte((byte)'S');
+			bes.WriteUInt16(classFile.AddInt((short)value.Value));
+		}
+		else if (value.ArgumentType == Types.Int32)
+		{
+			bes.WriteByte((byte)'I');
+			bes.WriteUInt16(classFile.AddInt((int)value.Value));
+		}
+		else if (value.ArgumentType == Types.Single)
+		{
+			bes.WriteByte((byte)'F');
+			bes.WriteUInt16(classFile.AddFloat((float)value.Value));
+		}
+		else if (value.ArgumentType == Types.Int64)
+		{
+			bes.WriteByte((byte)'J');
+			bes.WriteUInt16(classFile.AddLong((long)value.Value));
+		}
+		else if (value.ArgumentType == Types.Double)
+		{
+			bes.WriteByte((byte)'D');
+			bes.WriteUInt16(classFile.AddDouble((double)value.Value));
+		}
+		else if (value.ArgumentType == Types.String)
+		{
+			bes.WriteByte((byte)'s');
+			bes.WriteUInt16(classFile.AddUtf8((string)value.Value));
+		}
+		else if (value.ArgumentType == Types.Object.MakeArrayType())
+		{
+			CustomAttributeTypedArgument[] array = (CustomAttributeTypedArgument[])value.Value;
+			byte type = (byte)array[0].Value;
+			if (type == AnnotationDefaultAttribute.TAG_ARRAY)
+			{
+				bes.WriteByte((byte)'[');
+				bes.WriteUInt16((ushort)(array.Length - 1));
+				for (int i = 1; i < array.Length; i++)
+				{
+					WriteAnnotationElementValue(classFile, bes, array[i]);
+				}
+			}
+			else if (type == AnnotationDefaultAttribute.TAG_CLASS)
+			{
+				bes.WriteByte((byte)'c');
+				bes.WriteUInt16(classFile.AddUtf8((string)array[1].Value));
+			}
+			else if (type == AnnotationDefaultAttribute.TAG_ENUM)
+			{
+				bes.WriteByte((byte)'e');
+				bes.WriteUInt16(classFile.AddUtf8((string)array[1].Value));
+				bes.WriteUInt16(classFile.AddUtf8((string)array[2].Value));
+			}
+			else if (type == AnnotationDefaultAttribute.TAG_ANNOTATION)
+			{
+				bes.WriteByte((byte)'@');
+				bes.WriteUInt16(classFile.AddUtf8((string)array[1].Value));
+				bes.WriteUInt16((ushort)((array.Length - 2) / 2));
+				for (int i = 2; i < array.Length; i += 2)
+				{
+					bes.WriteUInt16(classFile.AddUtf8((string)array[i].Value));
+					WriteAnnotationElementValue(classFile, bes, array[i + 1]);
+				}
+			}
+			else
+			{
+				Console.Error.WriteLine("Warning: incorrect annotation default element tag: " + type);
+			}
+		}
+		else
+		{
+			Console.Error.WriteLine("Warning: incorrect annotation default element type: " + value.ArgumentType);
+		}
+	}
+
 	private static bool ExportNamespace(Type type)
 	{
 		if (namespaces.Count == 0)
@@ -642,10 +786,10 @@ static class NetExp
 		return false;
 	}
 
-	private static int ProcessAssembly(Assembly assembly, bool continueOnError)
+	private static int ProcessTypes(Type[] types, bool continueOnError)
 	{
 		int rc = 0;
-		foreach (Type t in assembly.GetTypes())
+		foreach (Type t in types)
 		{
 			if (t.IsPublic
 				&& ExportNamespace(t)
@@ -851,6 +995,9 @@ sealed class BootstrapBootstrapClassLoader : ClassLoaderWrapper
 		SetRemappedType(JVM.Import(typeof(string)), new StubTypeWrapper(Modifiers.Public | Modifiers.Final, "java.lang.String", javaLangObject, true));
 		SetRemappedType(JVM.Import(typeof(Exception)), new StubTypeWrapper(Modifiers.Public, "java.lang.Throwable", javaLangObject, true));
 		SetRemappedType(JVM.Import(typeof(IComparable)), new StubTypeWrapper(Modifiers.Public | Modifiers.Abstract | Modifiers.Interface, "java.lang.Comparable", null, true));
+		TypeWrapper tw = new StubTypeWrapper(Modifiers.Public | Modifiers.Abstract | Modifiers.Interface, "java.lang.AutoCloseable", null, true);
+		tw.SetMethods(new MethodWrapper[] { new SimpleCallMethodWrapper(tw, "close", "()V", JVM.Import(typeof(IDisposable)).GetMethod("Dispose"), PrimitiveTypeWrapper.VOID, TypeWrapper.EmptyArray, Modifiers.Public | Modifiers.Abstract, MemberFlags.None, SimpleOpCode.Callvirt, SimpleOpCode.Callvirt) });
+		SetRemappedType(JVM.Import(typeof(IDisposable)), tw);
 
 		RegisterInitiatingLoader(new StubTypeWrapper(Modifiers.Public, "java.lang.Enum", javaLangObject, false));
 		RegisterInitiatingLoader(new StubTypeWrapper(Modifiers.Public | Modifiers.Abstract | Modifiers.Interface, "java.lang.annotation.Annotation", null, false));
