@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2003, 2007, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 2003, 2008, Oracle and/or its affiliates. All rights reserved.
  * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
  * This code is free software; you can redistribute it and/or modify it
@@ -35,24 +35,23 @@ import java.util.concurrent.atomic.AtomicInteger;
 
 /**
  * Instances of the file descriptor class serve as an opaque handle
- * to the underlying machine-specific structure representing an open
- * file, an open socket, or another source or sink of bytes. The
- * main practical use for a file descriptor is to create a
- * <code>FileInputStream</code> or <code>FileOutputStream</code> to
- * contain it.
- * <p>
- * Applications should not create their own file descriptors.
+ * to the underlying machine-specific structure representing an
+ * open file, an open socket, or another source or sink of bytes.
+ * The main practical use for a file descriptor is to create a
+ * {@link FileInputStream} or {@link FileOutputStream} to contain it.
+ *
+ * <p>Applications should not create their own file descriptors.
  *
  * @author  Pavani Diwanji
- * @see     java.io.FileInputStream
- * @see     java.io.FileOutputStream
  * @since   JDK1.0
  */
 public final class FileDescriptor {
 
     private volatile cli.System.IO.Stream stream;
     private volatile cli.System.Net.Sockets.Socket socket;
-    
+    private volatile boolean nonBlockingSocket;
+    private volatile cli.System.IAsyncResult asyncResult;
+
     /**
      * HACK
      *   JRuby uses reflection to get at the handle (on Windows) and fd (on non-Windows)
@@ -138,7 +137,7 @@ public final class FileDescriptor {
     /**
      * A handle to the standard input stream. Usually, this file
      * descriptor is not used directly, but rather via the input stream
-     * known as <code>System.in</code>.
+     * known as {@code System.in}.
      *
      * @see     java.lang.System#in
      */
@@ -147,7 +146,7 @@ public final class FileDescriptor {
     /**
      * A handle to the standard output stream. Usually, this file
      * descriptor is not used directly, but rather via the output stream
-     * known as <code>System.out</code>.
+     * known as {@code System.out}.
      * @see     java.lang.System#out
      */
     public static final FileDescriptor out = standardStream(1);
@@ -155,7 +154,7 @@ public final class FileDescriptor {
     /**
      * A handle to the standard error stream. Usually, this file
      * descriptor is not used directly, but rather via the output stream
-     * known as <code>System.err</code>.
+     * known as {@code System.err}.
      *
      * @see     java.lang.System#err
      */
@@ -164,12 +163,12 @@ public final class FileDescriptor {
     /**
      * Tests if this file descriptor object is valid.
      *
-     * @return  <code>true</code> if the file descriptor object represents a
+     * @return  {@code true} if the file descriptor object represents a
      *          valid, open file, socket, or other active I/O connection;
-     *          <code>false</code> otherwise.
+     *          {@code false} otherwise.
      */
     public boolean valid() {
-        return stream != null;
+        return stream != null || socket != null;
     }
 
     /**
@@ -684,4 +683,48 @@ public final class FileDescriptor {
     {
         this.socket = socket;
     }
+
+    @ikvm.lang.Internal
+    public void setSocketBlocking(boolean blocking) throws IOException
+    {
+        this.nonBlockingSocket = !blocking;
+        try
+        {
+            if (false) throw new cli.System.Net.Sockets.SocketException();
+            if (false) throw new cli.System.ObjectDisposedException("");
+            socket.set_Blocking(blocking);
+        }
+        catch (cli.System.Net.Sockets.SocketException x)
+        {
+            if (x.get_ErrorCode() == java.net.SocketUtil.WSAEINVAL)
+            {
+                // Work around for winsock issue. You can't set a socket to blocking if a connection request is pending,
+                // so we'll have to set the blocking again in SocketChannelImpl.checkConnect().
+                return;
+            }
+            throw java.net.SocketUtil.convertSocketExceptionToIOException(x);
+        }
+        catch (cli.System.ObjectDisposedException _)
+        {
+            throw new java.net.SocketException("Socket is closed");
+        }
+    }
+    
+    @ikvm.lang.Internal
+    public boolean isSocketBlocking()
+    {
+        return !nonBlockingSocket;
+    }
+
+    @ikvm.lang.Internal
+    public cli.System.IAsyncResult getAsyncResult()
+    {
+        return asyncResult;
+    }
+
+    @ikvm.lang.Internal
+    public void setAsyncResult(cli.System.IAsyncResult asyncResult)
+    {
+        this.asyncResult = asyncResult;
+    }
 }
