@@ -1,5 +1,5 @@
 /* ZipFile.java --
-   Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2008
+   Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2008, 2011
    Free Software Foundation, Inc.
 
 This file is part of GNU Classpath.
@@ -39,21 +39,19 @@ exception statement from your version. */
 
 package java.util.zip;
 
-import gnu.java.util.EmptyEnumeration;
-
+import java.io.Closeable;
 import java.io.EOFException;
 import java.io.File;
-import java.io.FileNotFoundException;
 import java.io.IOException;
 import java.io.InputStream;
 import java.io.RandomAccessFile;
 import java.io.UnsupportedEncodingException;
-import java.nio.ByteBuffer;
 import java.nio.charset.Charset;
-import java.nio.charset.CharsetDecoder;
+import java.nio.charset.StandardCharsets;
 import java.util.Enumeration;
 import java.util.Iterator;
 import java.util.LinkedHashMap;
+import static java.util.zip.ZipConstants64.*;
 
 /**
  * This class represents a Zip archive.  You can ask for the contained
@@ -66,7 +64,7 @@ import java.util.LinkedHashMap;
  * @author Jochen Hoenicke
  * @author Artur Biesiadowski
  */
-public class ZipFile implements ZipConstants
+public class ZipFile implements ZipConstants, Closeable
 {
 
   /**
@@ -87,6 +85,9 @@ public class ZipFile implements ZipConstants
   // Name of this zip file.
   private final String name;
 
+  // Encoding to use for name and comment strings
+  private final Charset charset;
+
   // File from which zip entries are read.
   private final RandomAccessFile raf;
 
@@ -117,6 +118,17 @@ public class ZipFile implements ZipConstants
     this(file, OPEN_READ);
   }
 
+  /**
+   * Opens a Zip file reading the given File.
+   * @exception IOException if a i/o error occured.
+   * @exception ZipException if the file doesn't contain a valid zip
+   * archive.  
+   */
+  public ZipFile(File file, Charset charset) throws IOException
+  {
+    this(file, OPEN_READ, charset);
+  }
+
   /**
    * Opens a Zip file reading the given File in the given mode.
    *
@@ -133,14 +145,39 @@ public class ZipFile implements ZipConstants
    * @exception ZipException if the file doesn't contain a valid zip
    * archive.  
    */
-  public ZipFile(File file, int mode) throws ZipException, IOException
+  public ZipFile(File file, int mode) throws IOException
+  {
+    this(file, mode, StandardCharsets.UTF_8);
+  }
+
+  /**
+   * Opens a Zip file reading the given File in the given mode.
+   *
+   * If the OPEN_DELETE mode is specified, the zip file will be deleted at
+   * some time moment after it is opened. It will be deleted before the zip
+   * file is closed or the Virtual Machine exits.
+   * 
+   * The contents of the zip file will be accessible until it is closed.
+   *
+   * @since JDK1.7
+   * @param mode Must be one of OPEN_READ or OPEN_READ | OPEN_DELETE
+   * @param charset Character encoding to use for names and comments
+   *
+   * @exception IOException if a i/o error occured.
+   * @exception ZipException if the file doesn't contain a valid zip
+   * archive.  
+   */
+  public ZipFile(File file, int mode, Charset charset) throws IOException
   {
     if (mode != OPEN_READ && mode != (OPEN_READ | OPEN_DELETE))
       throw new IllegalArgumentException("Illegal mode: 0x" + Integer.toHexString(mode));
+    if (charset == null)
+      throw new NullPointerException("charset is null");
     if ((mode & OPEN_DELETE) != 0)
       file.deleteOnExit();
     this.raf = new RandomAccessFile(file, "r");
     this.name = file.getPath();
+    this.charset = charset;
 
     boolean valid = false;
 
@@ -185,29 +222,34 @@ public class ZipFile implements ZipConstants
    * @exception IOException if a i/o error occured.
    * @exception ZipException if the central directory is malformed 
    */
-  private void readEntries() throws ZipException, IOException
+  private void readEntries() throws IOException
   {
-    /* Search for the End Of Central Directory.  When a zip comment is 
-     * present the directory may start earlier.
-     * Note that a comment has a maximum length of 64K, so that is the
-     * maximum we search backwards.
-     */
     PartialInputStream inp = new PartialInputStream(4096);
-    long pos = raf.length() - ENDHDR;
-    long top = Math.max(0, pos - 65536);
-    do
-      {
-        if (pos < top)
-          throw new ZipException("error in opening zip file");
-        inp.seek(pos--);
-      }
-    while (inp.readLeInt() != ENDSIG);
-    
-    pos++;
+    long pos = inp.seekEndOfCentralDirectory();
     inp.skip(6);
-    int count = inp.readLeShort();
-    int centralSize = inp.readLeInt();    
-    int centralOffset = inp.readLeInt();
+    int count = inp.readLeUnsignedShort();
+    long centralSize = inp.readLeUnsignedInt();    
+    long centralOffset = inp.readLeUnsignedInt();
+
+    if (centralSize == ZIP64_MAGICVAL
+        || centralOffset == ZIP64_MAGICVAL
+        || count == ZIP64_MAGICCOUNT)
+      {
+        inp.seek(pos - ZIP64_LOCHDR);
+        if (inp.readLeInt() == ZIP64_LOCSIG)
+          {
+            inp.skip(4);
+            long zip64end = inp.readLeLong();
+            inp.seek(zip64end);
+            if (inp.readLeInt() == ZIP64_ENDSIG)
+              {
+                inp.skip(ZIP64_ENDSIZ - 4);
+                centralSize = inp.readLeLong();
+                centralOffset = inp.readLeLong();
+                pos = zip64end;
+              }
+          }
+      }
 
     if (centralSize > pos)
       throw new ZipException("invalid END header (bad central directory size)");
@@ -224,47 +266,96 @@ public class ZipFile implements ZipConstants
           throw new ZipException("invalid CEN header (bad signature)");
 
         inp.skip(4);
-        int flags = inp.readLeShort();
+        int flags = inp.readLeUnsignedShort();
         if ((flags & 1) != 0)
           throw new ZipException("invalid CEN header (encrypted entry)");
-        int method = inp.readLeShort();
+        int method = inp.readLeUnsignedShort();
         if (method != ZipEntry.STORED && method != ZipEntry.DEFLATED)
           throw new ZipException("invalid CEN header (bad compression method)");
-        int dostime = inp.readLeInt();
-        int crc = inp.readLeInt();
-        int csize = inp.readLeInt();
-        int size = inp.readLeInt();
-        int nameLen = inp.readLeShort();
-        int extraLen = inp.readLeShort();
-        int commentLen = inp.readLeShort();
+        ZipEntry entry = new ZipEntry();
+        entry.flag = flags;
+        entry.method = method;
+        entry.time = inp.readLeUnsignedInt();
+        entry.crc = inp.readLeUnsignedInt();
+        entry.csize = inp.readLeUnsignedInt();
+        entry.size = inp.readLeUnsignedInt();
+        int nameLen = inp.readLeUnsignedShort();
+        int extraLen = inp.readLeUnsignedShort();
+        int commentLen = inp.readLeUnsignedShort();
         inp.skip(8);
-        int offset = inp.readLeInt();
-        String name = inp.readString(nameLen);
+        entry.offset = inp.readLeUnsignedInt();
+        entry.name = inp.readString(nameLen, (flags & EFS) != 0);
 
-        ZipEntry entry = new ZipEntry(name);
-        entry.setMethod(method);
-        entry.setCrc(crc & 0xffffffffL);
-        entry.setSize(size & 0xffffffffL);
-        entry.setCompressedSize(csize & 0xffffffffL);
-        entry.time = dostime;
         if (extraLen > 0)
           {
             byte[] extra = new byte[extraLen];
             inp.readFully(extra);
-            entry.setExtra(extra);
+            entry.extra = extra;
+            readZip64ExtraField(entry, extra);
           }
         if (commentLen > 0)
           {
-            entry.setComment(inp.readString(commentLen));
+            entry.comment = inp.readString(commentLen, (flags & EFS) != 0);
           }
-        entry.offset = offset;
-        entries.put(name, entry);
+        entries.put(entry.name, entry);
       }
 
     if (inp.position() != pos)
       throw new ZipException("invalid CEN header (bad header size)");
   }
 
+  private static void readZip64ExtraField(ZipEntry entry, byte[] extra)
+  {
+    if (entry.csize == ZIP64_MAGICVAL || entry.size == ZIP64_MAGICVAL
+        || entry.offset == ZIP64_MAGICVAL)
+      {
+        for (int pos = 0; pos < extra.length - 4; )
+          {
+            int headerID = decodeLeUnsignedShort(extra, pos);
+            int dataSize = decodeLeUnsignedShort(extra, pos + 2);
+            pos += 4;
+            if (headerID == ZIP64_EXTID)
+              {
+                if (entry.size == ZIP64_MAGICVAL)
+                  {
+                    entry.size = decodeLeLong(extra, pos);
+                    pos += 8;
+                  }
+                if (entry.csize == ZIP64_MAGICVAL)
+                  {
+                    entry.csize = decodeLeLong(extra, pos);
+                    pos += 8;
+                  }
+                if (entry.offset == ZIP64_MAGICVAL)
+                  {
+                    entry.offset = decodeLeLong(extra, pos);
+                    pos += 8;
+                  }
+                break;
+              }
+            pos += dataSize;
+          }
+      }
+  }
+
+  private static int decodeLeUnsignedShort(byte[] b, int off)
+  {
+    return (b[off] & 0xFF) | ((b[off + 1] & 0xFF) << 8);
+  }
+
+  private static long decodeLeLong(byte[] b, int off)
+  {
+    return 0
+        | ((b[off + 0] & 0xFFL) << 0)
+        | ((b[off + 1] & 0xFFL) << 8)
+        | ((b[off + 2] & 0xFFL) << 16)
+        | ((b[off + 3] & 0xFFL) << 24)
+        | ((b[off + 4] & 0xFFL) << 32)
+        | ((b[off + 5] & 0xFFL) << 40)
+        | ((b[off + 6] & 0xFFL) << 48)
+        | ((b[off + 7] & 0xFFL) << 56);
+  }
+
   /**
    * Closes the ZipFile.  This also closes all input streams given by
    * this class.  After this is called, no further method should be
@@ -322,7 +413,7 @@ public class ZipFile implements ZipConstants
     // If we didn't find it, maybe it's a directory.
     if (entry == null && !name.endsWith("/"))
       entry = entries.get(name + '/');
-    return entry != null ? new ZipEntry(entry, name) : null;
+    return entry != null ? (ZipEntry)entry.clone() : null;
   }
 
   /**
@@ -364,8 +455,8 @@ public class ZipFile implements ZipConstants
 
             skip(22);
 
-            int nameLen = readLeShort();
-            int extraLen = readLeShort();
+            int nameLen = readLeUnsignedShort();
+            int extraLen = readLeUnsignedShort();
             skip(nameLen + extraLen);
 
             setLength(zipEntry.getCompressedSize());
@@ -422,7 +513,31 @@ public class ZipFile implements ZipConstants
     checkClosed();
     return entries.size();
   }
-  
+
+  /**
+   * Returns the zip file comment.
+   *
+   * @exception IllegalStateException when the ZipFile has already been closed
+   */
+  public synchronized String getComment()
+  {
+    checkClosed();
+    try
+      {
+        PartialInputStream inp = new PartialInputStream(4096);
+        long pos = inp.seekEndOfCentralDirectory();
+        inp.skip(16);
+        int commentLength = inp.readLeUnsignedShort();
+        if (commentLength == 0)
+          return null;
+        return inp.readString(commentLength, false);
+      }
+    catch (IOException _)
+      {
+        return null;
+      }
+  }
+
   private static class ZipEntryEnumeration implements Enumeration<ZipEntry>
   {
     private final Iterator<ZipEntry> elements;
@@ -469,6 +584,31 @@ public class ZipFile implements ZipConstants
       lazy = true;
     }
 
+    // Seek to the "end of central directory record" and position
+    // the stream after the ENDSIG and return the file offset
+    // where the record starts.
+    long seekEndOfCentralDirectory() throws IOException
+    {
+      /* Search for the End Of Central Directory.  When a zip comment is 
+       * present the directory may start earlier.
+       * Note that a comment has a maximum length of 64K, so that is the
+       * maximum we search backwards.
+       */
+      long length = raf.length();
+      if (length == 0)
+        throw new ZipException("zip file is empty");
+      long pos = length - ENDHDR;
+      long top = Math.max(0, pos - 65536);
+      do
+        {
+          if (pos < top)
+            throw new ZipException("error in opening zip file");
+          seek(pos--);
+        }
+      while (readLeInt() != ENDSIG);
+      return pos + 1;
+    }
+
     void lazyInitialSeek() throws IOException
     {
     }
@@ -616,7 +756,7 @@ public class ZipFile implements ZipConstants
         throw new EOFException();
     }
 
-    int readLeShort() throws IOException
+    int readLeUnsignedShort() throws IOException
     {
       int result;
       if(pos + 1 < buffer.length)
@@ -659,8 +799,18 @@ public class ZipFile implements ZipConstants
       return result;
     }
 
+    final long readLeUnsignedInt() throws IOException
+    {
+      return readLeInt() & 0xffffffffL;
+    }
+    
+    final long readLeLong() throws IOException
+    {
+      return readLeUnsignedInt() | (readLeUnsignedInt() << 32);
+    }
+
     /**
-     * Decode chars from byte buffer using UTF8 encoding.  This
+     * Decode chars from byte buffer using charset encoding.  This
      * operation is performance-critical since a jar file contains a
      * large number of strings for the name of each file in the
      * archive.  This routine therefore avoids using the expensive
@@ -675,9 +825,12 @@ public class ZipFile implements ZipConstants
      *
      * @return a String that contains the decoded characters.
      */
-    private String decodeChars(byte[] buffer, int pos, int length)
+    private String decodeChars(byte[] buffer, int pos, int length, boolean utf8)
       throws IOException
     {
+      if (!utf8 && charset != StandardCharsets.UTF_8)
+        return new String(buffer, pos, length, charset);
+
       for (int i = pos; i < pos + length; i++)
         {
           if (buffer[i] <= 0)
@@ -686,7 +839,7 @@ public class ZipFile implements ZipConstants
       return new String(buffer, 0, pos, length);
     }
 
-    String readString(int length) throws IOException
+    String readString(int length, boolean utf8) throws IOException
     {
       if (length > end - (bufferOffset + pos))
         throw new EOFException();
@@ -696,14 +849,14 @@ public class ZipFile implements ZipConstants
         {
           if (buffer.length - pos >= length)
             {
-              result = decodeChars(buffer, pos, length);
+              result = decodeChars(buffer, pos, length, utf8);
               pos += length;
             }
           else
             {
               byte[] b = new byte[length];
               readFully(b);
-              result = decodeChars(b, 0, length);
+              result = decodeChars(b, 0, length, utf8);
             }
         }
       catch (UnsupportedEncodingException uee)
