@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2006, 2007, 2009 Jeroen Frijters
+  Copyright (C) 2006-2011 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -32,11 +32,13 @@ import ikvm.lang.Internal;
 import java.lang.reflect.Field;
 import java.lang.reflect.Modifier;
 import java.lang.reflect.ReflectHelper;
+import java.security.ProtectionDomain;
 import java.util.ArrayList;
 
 public final class Unsafe
 {
     public static final int INVALID_FIELD_OFFSET = -1;
+    public static final int ARRAY_BYTE_BASE_OFFSET = 0;
     // NOTE sun.corba.Bridge actually access this field directly (via reflection),
     // so the name must match the JDK name.
     private static final Unsafe theUnsafe = new Unsafe();
@@ -890,6 +892,102 @@ public final class Unsafe
 	    putByte(destAddress++, getByte(srcAddress++));
 	}
     }
+    
+    @SecurityPermissionAttribute.Annotation(value = SecurityAction.__Enum.LinkDemand, UnmanagedCode = true)
+    @cli.System.Security.SecurityCriticalAttribute.Annotation
+    public void copyMemory(Object srcBase, long srcOffset, Object destBase, long destOffset, long bytes)
+    {
+        if (srcBase == null)
+        {
+            if (destBase instanceof byte[])
+            {
+                cli.System.Runtime.InteropServices.Marshal.Copy(IntPtr.op_Explicit(srcOffset), (byte[])destBase, (int)destOffset, (int)bytes);
+            }
+            else if (destBase instanceof boolean[])
+            {
+                byte[] tmp = new byte[(int)bytes];
+                copyMemory(srcBase, srcOffset, tmp, 0, bytes);
+                copyMemory(tmp, 0, destBase, destOffset, bytes);
+            }
+            else if (destBase instanceof short[])
+            {
+                cli.System.Runtime.InteropServices.Marshal.Copy(IntPtr.op_Explicit(srcOffset), (short[])destBase, (int)(destOffset >> 1), (int)(bytes >> 1));
+            }
+            else if (destBase instanceof char[])
+            {
+                cli.System.Runtime.InteropServices.Marshal.Copy(IntPtr.op_Explicit(srcOffset), (char[])destBase, (int)(destOffset >> 1), (int)(bytes >> 1));
+            }
+            else if (destBase instanceof int[])
+            {
+                cli.System.Runtime.InteropServices.Marshal.Copy(IntPtr.op_Explicit(srcOffset), (int[])destBase, (int)(destOffset >> 2), (int)(bytes >> 2));
+            }
+            else if (destBase instanceof float[])
+            {
+                cli.System.Runtime.InteropServices.Marshal.Copy(IntPtr.op_Explicit(srcOffset), (float[])destBase, (int)(destOffset >> 2), (int)(bytes >> 2));
+            }
+            else if (destBase instanceof long[])
+            {
+                cli.System.Runtime.InteropServices.Marshal.Copy(IntPtr.op_Explicit(srcOffset), (long[])destBase, (int)(destOffset >> 3), (int)(bytes >> 3));
+            }
+            else if (destBase instanceof double[])
+            {
+                cli.System.Runtime.InteropServices.Marshal.Copy(IntPtr.op_Explicit(srcOffset), (double[])destBase, (int)(destOffset >> 3), (int)(bytes >> 3));
+            }
+            else if (destBase == null)
+            {
+                copyMemory(srcOffset, destOffset, bytes);
+            }
+            else
+            {
+                throw new IllegalArgumentException();
+            }
+        }
+        else if (srcBase instanceof cli.System.Array && destBase instanceof cli.System.Array)
+        {
+            cli.System.Buffer.BlockCopy((cli.System.Array)srcBase, (int)srcOffset, (cli.System.Array)destBase, (int)destOffset, (int)bytes);
+        }
+        else
+        {
+            if (srcBase instanceof byte[])
+            {
+                cli.System.Runtime.InteropServices.Marshal.Copy((byte[])srcBase, (int)srcOffset, IntPtr.op_Explicit(destOffset), (int)bytes);
+            }
+            else if (srcBase instanceof boolean[])
+            {
+                byte[] tmp = new byte[(int)bytes];
+                copyMemory(srcBase, srcOffset, tmp, 0, bytes);
+                copyMemory(tmp, 0, destBase, destOffset, bytes);
+            }
+            else if (srcBase instanceof short[])
+            {
+                cli.System.Runtime.InteropServices.Marshal.Copy((short[])srcBase, (int)(srcOffset >> 1), IntPtr.op_Explicit(destOffset), (int)(bytes >> 1));
+            }
+            else if (srcBase instanceof char[])
+            {
+                cli.System.Runtime.InteropServices.Marshal.Copy((char[])srcBase, (int)(srcOffset >> 1), IntPtr.op_Explicit(destOffset), (int)(bytes >> 1));
+            }
+            else if (srcBase instanceof int[])
+            {
+                cli.System.Runtime.InteropServices.Marshal.Copy((int[])srcBase, (int)(srcOffset >> 2), IntPtr.op_Explicit(destOffset), (int)(bytes >> 2));
+            }
+            else if (srcBase instanceof float[])
+            {
+                cli.System.Runtime.InteropServices.Marshal.Copy((float[])srcBase, (int)(srcOffset >> 2), IntPtr.op_Explicit(destOffset), (int)(bytes >> 2));
+            }
+            else if (srcBase instanceof long[])
+            {
+                cli.System.Runtime.InteropServices.Marshal.Copy((long[])srcBase, (int)(srcOffset >> 3), IntPtr.op_Explicit(destOffset), (int)(bytes >> 3));
+            }
+            else if (srcBase instanceof double[])
+            {
+                cli.System.Runtime.InteropServices.Marshal.Copy((double[])srcBase, (int)(srcOffset >> 3), IntPtr.op_Explicit(destOffset), (int)(bytes >> 3));
+            }
+            else
+            {
+                throw new IllegalArgumentException();
+            }
+        }
+    }
 
     @SecurityPermissionAttribute.Annotation(value = SecurityAction.__Enum.LinkDemand, UnmanagedCode = true)
     @cli.System.Security.SecurityCriticalAttribute.Annotation
@@ -1007,4 +1105,16 @@ public final class Unsafe
     {
         return -1;
     }
+
+    public void unpark(Object thread)
+    {
+        java.util.concurrent.locks.LockSupport.unpark((Thread)thread);
+    }
+
+    public Object staticFieldBase(Field f)
+    {
+        return null;
+    }
+
+    public native Class defineClass(String name, byte[] buf, int offset, int length, ClassLoader cl, ProtectionDomain pd);
 }
