@@ -34,28 +34,90 @@ namespace IKVM.Reflection
 	public sealed class CustomAttributeData
 	{
 		internal static readonly IList<CustomAttributeData> EmptyList = new List<CustomAttributeData>(0).AsReadOnly();
-		private Module module;
-		private int index;
+
+		/*
+		 * There are several states a CustomAttributeData object can be in:
+		 * 
+		 * 1) Unresolved Custom Attribute
+		 *    - index >= 0
+		 *    - declSecurityBlob == null
+		 *    - lazyConstructor = null
+		 *    - lazyConstructorArguments = null
+		 *    - lazyNamedArguments = null
+		 * 
+		 * 2) Resolved Custom Attribute
+		 *    - index >= 0
+		 *    - declSecurityBlob == null
+		 *    - lazyConstructor != null
+		 *    - lazyConstructorArguments != null
+		 *    - lazyNamedArguments != null
+		 *    
+		 * 3) Pre-resolved Custom Attribute
+		 *    - index = -1
+		 *    - declSecurityBlob == null
+		 *    - lazyConstructor != null
+		 *    - lazyConstructorArguments != null
+		 *    - lazyNamedArguments != null
+		 *    
+		 * 4) Pseudo Custom Attribute, .NET 1.x declarative security or result of CustomAttributeBuilder.ToData()
+		 *    - index = -1
+		 *    - declSecurityBlob == null
+		 *    - lazyConstructor != null
+		 *    - lazyConstructorArguments != null
+		 *    - lazyNamedArguments != null
+		 *    
+		 * 5) Unresolved declarative security
+		 *    - index = -1
+		 *    - declSecurityBlob != null
+		 *    - lazyConstructor != null
+		 *    - lazyConstructorArguments != null
+		 *    - lazyNamedArguments == null
+		 * 
+		 * 6) Resolved declarative security
+		 *    - index = -1
+		 *    - declSecurityBlob == null
+		 *    - lazyConstructor != null
+		 *    - lazyConstructorArguments != null
+		 *    - lazyNamedArguments != null
+		 * 
+		 */
+		private readonly Module module;
+		private readonly int index;
+		private readonly byte[] declSecurityBlob;
 		private ConstructorInfo lazyConstructor;
 		private IList<CustomAttributeTypedArgument> lazyConstructorArguments;
 		private IList<CustomAttributeNamedArgument> lazyNamedArguments;
 
+		// 1) Unresolved Custom Attribute
 		internal CustomAttributeData(Module module, int index)
 		{
 			this.module = module;
 			this.index = index;
 		}
 
-		internal CustomAttributeData(ConstructorInfo constructor, object[] args, List<CustomAttributeNamedArgument> namedArguments)
+		// 4) Pseudo Custom Attribute, .NET 1.x declarative security
+		internal CustomAttributeData(Module module, ConstructorInfo constructor, object[] args, List<CustomAttributeNamedArgument> namedArguments)
+			: this(module, constructor, WrapConstructorArgs(args, constructor.MethodSignature), namedArguments)
+		{
+		}
+
+		private static List<CustomAttributeTypedArgument> WrapConstructorArgs(object[] args, MethodSignature sig)
 		{
-			this.lazyConstructor = constructor;
-			MethodSignature sig = constructor.MethodSignature;
 			List<CustomAttributeTypedArgument> list = new List<CustomAttributeTypedArgument>();
 			for (int i = 0; i < args.Length; i++)
 			{
 				list.Add(new CustomAttributeTypedArgument(sig.GetParameterType(i), args[i]));
 			}
-			lazyConstructorArguments = list.AsReadOnly();
+			return list;
+		}
+
+		// 4) Pseudo Custom Attribute, .NET 1.x declarative security or result of CustomAttributeBuilder.ToData()
+		internal CustomAttributeData(Module module, ConstructorInfo constructor, List<CustomAttributeTypedArgument> constructorArgs, List<CustomAttributeNamedArgument> namedArguments)
+		{
+			this.module = module;
+			this.index = -1;
+			this.lazyConstructor = constructor;
+			lazyConstructorArguments = constructorArgs.AsReadOnly();
 			if (namedArguments == null)
 			{
 				this.lazyNamedArguments = Empty<CustomAttributeNamedArgument>.Array;
@@ -66,8 +128,11 @@ namespace IKVM.Reflection
 			}
 		}
 
+		// 3) Pre-resolved Custom Attribute
 		internal CustomAttributeData(Assembly asm, ConstructorInfo constructor, ByteReader br)
 		{
+			this.module = asm.ManifestModule;
+			this.index = -1;
 			this.lazyConstructor = constructor;
 			if (br.Length == 0)
 			{
@@ -93,11 +158,13 @@ namespace IKVM.Reflection
 			sb.Append(Constructor.DeclaringType.FullName);
 			sb.Append('(');
 			string sep = "";
-			foreach (CustomAttributeTypedArgument arg in ConstructorArguments)
+			ParameterInfo[] parameters = Constructor.GetParameters();
+			IList<CustomAttributeTypedArgument> args = ConstructorArguments;
+			for (int i = 0; i < parameters.Length; i++)
 			{
 				sb.Append(sep);
 				sep = ", ";
-				AppendValue(sb, arg);
+				AppendValue(sb, parameters[i].ParameterType, args[i]);
 			}
 			foreach (CustomAttributeNamedArgument named in NamedArguments)
 			{
@@ -105,22 +172,49 @@ namespace IKVM.Reflection
 				sep = ", ";
 				sb.Append(named.MemberInfo.Name);
 				sb.Append(" = ");
-				AppendValue(sb, named.TypedValue);
+				FieldInfo fi = named.MemberInfo as FieldInfo;
+				Type type = fi != null ? fi.FieldType : ((PropertyInfo)named.MemberInfo).PropertyType;
+				AppendValue(sb, type, named.TypedValue);
 			}
 			sb.Append(')');
 			sb.Append(']');
 			return sb.ToString();
 		}
 
-		private static void AppendValue(StringBuilder sb, CustomAttributeTypedArgument arg)
+		private static void AppendValue(StringBuilder sb, Type type, CustomAttributeTypedArgument arg)
 		{
 			if (arg.ArgumentType == arg.ArgumentType.Module.universe.System_String)
 			{
 				sb.Append('"').Append(arg.Value).Append('"');
 			}
+			else if (arg.ArgumentType.IsArray)
+			{
+				Type elementType = arg.ArgumentType.GetElementType();
+				string elementTypeName;
+				if (elementType.IsPrimitive
+					|| elementType == type.Module.universe.System_Object
+					|| elementType == type.Module.universe.System_String
+					|| elementType == type.Module.universe.System_Type)
+				{
+					elementTypeName = elementType.Name;
+				}
+				else
+				{
+					elementTypeName = elementType.FullName;
+				}
+				sb.Append("new ").Append(elementTypeName).Append("[").Append(((Array)arg.Value).Length).Append("] { ");
+				string sep = "";
+				foreach (CustomAttributeTypedArgument elem in (CustomAttributeTypedArgument[])arg.Value)
+				{
+					sb.Append(sep);
+					sep = ", ";
+					AppendValue(sb, elementType, elem);
+				}
+				sb.Append(" }");
+			}
 			else
 			{
-				if (arg.ArgumentType.IsEnum)
+				if (arg.ArgumentType != type)
 				{
 					sb.Append('(');
 					sb.Append(arg.ArgumentType.FullName);
@@ -140,19 +234,10 @@ namespace IKVM.Reflection
 				for (int j = 0; j < count; j++)
 				{
 					Type type = ReadType(asm, br);
-					ConstructorInfo constructor;
-					if (type == u.System_Security_Permissions_HostProtectionAttribute && action == (int)System.Security.Permissions.SecurityAction.LinkDemand)
-					{
-						constructor = type.GetConstructor(BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic, null, Type.EmptyTypes, null);
-					}
-					else
-					{
-						constructor = type.GetConstructor(BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic, null, new Type[] { u.System_Security_Permissions_SecurityAction }, null);
-					}
+					ConstructorInfo constructor = type.GetPseudoCustomAttributeConstructor(u.System_Security_Permissions_SecurityAction);
 					// LAMESPEC there is an additional length here (probably of the named argument list)
-					ByteReader slice = br.Slice(br.ReadCompressedInt());
-					// LAMESPEC the count of named arguments is a compressed integer (instead of UInt16 as NumNamed in custom attributes)
-					list.Add(new CustomAttributeData(constructor, action, ReadNamedArguments(asm, slice, slice.ReadCompressedInt(), type)));
+					byte[] blob = br.ReadBytes(br.ReadCompressedInt());
+					list.Add(new CustomAttributeData(asm, constructor, action, blob));
 				}
 			}
 			else
@@ -164,22 +249,25 @@ namespace IKVM.Reflection
 					buf[i] = br.ReadChar();
 				}
 				string xml = new String(buf);
-				ConstructorInfo constructor = u.System_Security_Permissions_PermissionSetAttribute.GetConstructor(new Type[] { u.System_Security_Permissions_SecurityAction });
+				ConstructorInfo constructor = u.System_Security_Permissions_PermissionSetAttribute.GetPseudoCustomAttributeConstructor(u.System_Security_Permissions_SecurityAction);
 				List<CustomAttributeNamedArgument> args = new List<CustomAttributeNamedArgument>();
-				args.Add(new CustomAttributeNamedArgument(u.System_Security_Permissions_PermissionSetAttribute.GetProperty("XML"),
+				args.Add(new CustomAttributeNamedArgument(GetProperty(u.System_Security_Permissions_PermissionSetAttribute, "XML", u.System_String),
 					new CustomAttributeTypedArgument(u.System_String, xml)));
-				list.Add(new CustomAttributeData(constructor, action, args));
+				list.Add(new CustomAttributeData(asm.ManifestModule, constructor, new object[] { action }, args));
 			}
 		}
 
-		private CustomAttributeData(ConstructorInfo constructor, int securityAction, IList<CustomAttributeNamedArgument> namedArguments)
+		// 5) Unresolved declarative security
+		internal CustomAttributeData(Assembly asm, ConstructorInfo constructor, int securityAction, byte[] blob)
 		{
+			this.module = asm.ManifestModule;
+			this.index = -1;
 			Universe u = constructor.Module.universe;
 			this.lazyConstructor = constructor;
 			List<CustomAttributeTypedArgument> list = new List<CustomAttributeTypedArgument>();
 			list.Add(new CustomAttributeTypedArgument(u.System_Security_Permissions_SecurityAction, securityAction));
 			this.lazyConstructorArguments =  list.AsReadOnly();
-			this.lazyNamedArguments = namedArguments;
+			this.declSecurityBlob = blob;
 		}
 
 		private static Type ReadFieldOrPropType(Assembly asm, ByteReader br)
@@ -222,7 +310,7 @@ namespace IKVM.Reflection
 				case 0x51:
 					return u.System_Object;
 				default:
-					throw new InvalidOperationException();
+					throw new BadImageFormatException();
 			}
 		}
 
@@ -233,6 +321,54 @@ namespace IKVM.Reflection
 			{
 				return new CustomAttributeTypedArgument(type, br.ReadString());
 			}
+			else if (type == u.System_Boolean)
+			{
+				return new CustomAttributeTypedArgument(type, br.ReadByte() != 0);
+			}
+			else if (type == u.System_Char)
+			{
+				return new CustomAttributeTypedArgument(type, br.ReadChar());
+			}
+			else if (type == u.System_Single)
+			{
+				return new CustomAttributeTypedArgument(type, br.ReadSingle());
+			}
+			else if (type == u.System_Double)
+			{
+				return new CustomAttributeTypedArgument(type, br.ReadDouble());
+			}
+			else if (type == u.System_SByte)
+			{
+				return new CustomAttributeTypedArgument(type, br.ReadSByte());
+			}
+			else if (type == u.System_Int16)
+			{
+				return new CustomAttributeTypedArgument(type, br.ReadInt16());
+			}
+			else if (type == u.System_Int32)
+			{
+				return new CustomAttributeTypedArgument(type, br.ReadInt32());
+			}
+			else if (type == u.System_Int64)
+			{
+				return new CustomAttributeTypedArgument(type, br.ReadInt64());
+			}
+			else if (type == u.System_Byte)
+			{
+				return new CustomAttributeTypedArgument(type, br.ReadByte());
+			}
+			else if (type == u.System_UInt16)
+			{
+				return new CustomAttributeTypedArgument(type, br.ReadUInt16());
+			}
+			else if (type == u.System_UInt32)
+			{
+				return new CustomAttributeTypedArgument(type, br.ReadUInt32());
+			}
+			else if (type == u.System_UInt64)
+			{
+				return new CustomAttributeTypedArgument(type, br.ReadUInt64());
+			}
 			else if (type == u.System_Type)
 			{
 				return new CustomAttributeTypedArgument(type, ReadType(asm, br));
@@ -262,35 +398,7 @@ namespace IKVM.Reflection
 			}
 			else
 			{
-				switch (Type.GetTypeCode(type))
-				{
-					case TypeCode.Boolean:
-						return new CustomAttributeTypedArgument(type, br.ReadByte() != 0);
-					case TypeCode.Char:
-						return new CustomAttributeTypedArgument(type, br.ReadChar());
-					case TypeCode.Single:
-						return new CustomAttributeTypedArgument(type, br.ReadSingle());
-					case TypeCode.Double:
-						return new CustomAttributeTypedArgument(type, br.ReadDouble());
-					case TypeCode.SByte:
-						return new CustomAttributeTypedArgument(type, br.ReadSByte());
-					case TypeCode.Int16:
-						return new CustomAttributeTypedArgument(type, br.ReadInt16());
-					case TypeCode.Int32:
-						return new CustomAttributeTypedArgument(type, br.ReadInt32());
-					case TypeCode.Int64:
-						return new CustomAttributeTypedArgument(type, br.ReadInt64());
-					case TypeCode.Byte:
-						return new CustomAttributeTypedArgument(type, br.ReadByte());
-					case TypeCode.UInt16:
-						return new CustomAttributeTypedArgument(type, br.ReadUInt16());
-					case TypeCode.UInt32:
-						return new CustomAttributeTypedArgument(type, br.ReadUInt32());
-					case TypeCode.UInt64:
-						return new CustomAttributeTypedArgument(type, br.ReadUInt64());
-					default:
-						throw new InvalidOperationException();
-				}
+				throw new InvalidOperationException();
 			}
 		}
 
@@ -306,7 +414,7 @@ namespace IKVM.Reflection
 				// there are broken compilers that emit an extra NUL character after the type name
 				typeName = typeName.Substring(0, typeName.Length - 1);
 			}
-			return asm.universe.GetType(asm, typeName, true);
+			return TypeNameParser.Parse(typeName, true).GetType(asm.universe, asm, true, typeName, true);
 		}
 
 		private static IList<CustomAttributeTypedArgument> ReadConstructorArguments(Assembly asm, ByteReader br, ConstructorInfo constructor)
@@ -334,26 +442,23 @@ namespace IKVM.Reflection
 				switch (fieldOrProperty)
 				{
 					case 0x53:
-						member = GetField(type, name);
+						member = GetField(type, name, fieldOrPropertyType);
 						break;
 					case 0x54:
-						member = GetProperty(type, name);
+						member = GetProperty(type, name, fieldOrPropertyType);
 						break;
 					default:
 						throw new BadImageFormatException();
 				}
-				if (member == null)
-				{
-					throw new BadImageFormatException();
-				}
 				list.Add(new CustomAttributeNamedArgument(member, value));
 			}
 			return list.AsReadOnly();
 		}
 
-		private static FieldInfo GetField(Type type, string name)
+		private static FieldInfo GetField(Type type, string name, Type fieldType)
 		{
-			for (; type != null; type = type.BaseType)
+			Type org = type;
+			for (; type != null && !type.__IsMissing; type = type.BaseType)
 			{
 				foreach (FieldInfo field in type.__GetDeclaredFields())
 				{
@@ -363,12 +468,20 @@ namespace IKVM.Reflection
 					}
 				}
 			}
-			return null;
+			// if the field is missing, we stick the missing field on the first missing base type
+			if (type == null)
+			{
+				type = org;
+			}
+			FieldSignature sig = FieldSignature.Create(fieldType, null, null);
+			return type.FindField(name, sig)
+				?? type.Module.universe.GetMissingFieldOrThrow(type, name, sig);
 		}
 
-		private static PropertyInfo GetProperty(Type type, string name)
+		private static PropertyInfo GetProperty(Type type, string name, Type propertyType)
 		{
-			for (; type != null; type = type.BaseType)
+			Type org = type;
+			for (; type != null && !type.__IsMissing; type = type.BaseType)
 			{
 				foreach (PropertyInfo property in type.__GetDeclaredProperties())
 				{
@@ -378,7 +491,12 @@ namespace IKVM.Reflection
 					}
 				}
 			}
-			return null;
+			// if the property is missing, we stick the missing property on the first missing base type
+			if (type == null)
+			{
+				type = org;
+			}
+			return type.Module.universe.GetMissingPropertyOrThrow(type, name, PropertySignature.Create(CallingConventions.Standard | CallingConventions.HasThis, propertyType, null, null, null, null, null));
 		}
 
 		[Obsolete("Use Constructor.DeclaringType instead.")]
@@ -397,7 +515,18 @@ namespace IKVM.Reflection
 
 		public byte[] __GetBlob()
 		{
-			return ((ModuleReader)module).GetBlobCopy(module.CustomAttribute.records[index].Value);
+			if (declSecurityBlob != null)
+			{
+				return (byte[])declSecurityBlob.Clone();
+			}
+			else if (index == -1)
+			{
+				return __ToBuilder().GetBlob(module.Assembly);
+			}
+			else
+			{
+				return ((ModuleReader)module).GetBlobCopy(module.CustomAttribute.records[index].Value);
+			}
 		}
 
 		public ConstructorInfo Constructor
@@ -430,7 +559,18 @@ namespace IKVM.Reflection
 			{
 				if (lazyNamedArguments == null)
 				{
-					LazyParseArguments();
+					if (index >= 0)
+					{
+						// 1) Unresolved Custom Attribute
+						LazyParseArguments();
+					}
+					else
+					{
+						// 5) Unresolved declarative security
+						ByteReader br = new ByteReader(declSecurityBlob, 0, declSecurityBlob.Length);
+						// LAMESPEC the count of named arguments is a compressed integer (instead of UInt16 as NumNamed in custom attributes)
+						lazyNamedArguments = ReadNamedArguments(module.Assembly, br, br.ReadCompressedInt(), Constructor.DeclaringType);
+					}
 				}
 				return lazyNamedArguments;
 			}
@@ -458,10 +598,11 @@ namespace IKVM.Reflection
 
 		public CustomAttributeBuilder __ToBuilder()
 		{
+			ParameterInfo[] parameters = Constructor.GetParameters();
 			object[] args = new object[ConstructorArguments.Count];
 			for (int i = 0; i < args.Length; i++)
 			{
-				args[i] = RewrapArray(ConstructorArguments[i]);
+				args[i] = RewrapArray(parameters[i].ParameterType, ConstructorArguments[i]);
 			}
 			List<PropertyInfo> namedProperties = new List<PropertyInfo>();
 			List<object> propertyValues = new List<object>();
@@ -469,30 +610,36 @@ namespace IKVM.Reflection
 			List<object> fieldValues = new List<object>();
 			foreach (CustomAttributeNamedArgument named in NamedArguments)
 			{
-				if (named.MemberInfo is PropertyInfo)
+				PropertyInfo pi = named.MemberInfo as PropertyInfo;
+				if (pi != null)
 				{
-					namedProperties.Add((PropertyInfo)named.MemberInfo);
-					propertyValues.Add(RewrapArray(named.TypedValue));
+					namedProperties.Add(pi);
+					propertyValues.Add(RewrapArray(pi.PropertyType, named.TypedValue));
 				}
 				else
 				{
-					namedFields.Add((FieldInfo)named.MemberInfo);
-					fieldValues.Add(RewrapArray(named.TypedValue));
+					FieldInfo fi = (FieldInfo)named.MemberInfo;
+					namedFields.Add(fi);
+					fieldValues.Add(RewrapArray(fi.FieldType, named.TypedValue));
 				}
 			}
 			return new CustomAttributeBuilder(Constructor, args, namedProperties.ToArray(), propertyValues.ToArray(), namedFields.ToArray(), fieldValues.ToArray());
 		}
 
-		private static object RewrapArray(CustomAttributeTypedArgument arg)
+		private static object RewrapArray(Type type, CustomAttributeTypedArgument arg)
 		{
 			IList<CustomAttributeTypedArgument> list = arg.Value as IList<CustomAttributeTypedArgument>;
 			if (list != null)
 			{
+				Type elementType = arg.ArgumentType.GetElementType();
 				object[] arr = new object[list.Count];
 				for (int i = 0; i < arr.Length; i++)
 				{
-					// note that CLI spec only allows one dimensional arrays, so we don't need to rewrap the elements
-					arr[i] = list[i].Value;
+					arr[i] = RewrapArray(elementType, list[i]);
+				}
+				if (type == type.Module.universe.System_Object)
+				{
+					return CustomAttributeBuilder.__MakeTypedArgument(arg.ArgumentType, arr);
 				}
 				return arr;
 			}
@@ -574,6 +721,11 @@ namespace IKVM.Reflection
 			}
 		}
 
+		public static IList<CustomAttributeData> __GetCustomAttributes(Type type, Type interfaceType, Type attributeType, bool inherit)
+		{
+			return type.GetInterfaceImplCustomAttributes(interfaceType, attributeType);
+		}
+
 		public static IList<CustomAttributeData> __GetDeclarativeSecurity(Assembly assembly)
 		{
 			if (assembly.__IsMissing)
