@@ -160,6 +160,7 @@ namespace IKVM.Reflection.Emit
 		private readonly List<ExceptionBlock> exceptions = new List<ExceptionBlock>();
 		private readonly Stack<ExceptionBlock> exceptionStack = new Stack<ExceptionBlock>();
 		private ushort maxStack;
+		private bool fatHeader;
 		private int stackHeight;
 		private Scope scope;
 		private byte exceptionBlockAssistanceMode = EBAM_COMPAT;
@@ -266,6 +267,24 @@ namespace IKVM.Reflection.Emit
 			exceptionBlockAssistanceMode = EBAM_CLEVER;
 		}
 
+		// non-standard API
+		public int __MaxStackSize
+		{
+			get { return maxStack; }
+			set
+			{
+				maxStack = (ushort)value;
+				fatHeader = true;
+			}
+		}
+
+		// non-standard API
+		// returns -1 if the current position is currently unreachable
+		public int __StackHeight
+		{
+			get { return stackHeight; }
+		}
+
 		// new in .NET 4.0
 		public int ILOffset
 		{
@@ -918,7 +937,7 @@ namespace IKVM.Reflection.Emit
 			int localVarSigTok = 0;
 
 			int rva;
-			if (locals.Count == 0 && exceptions.Count == 0 && maxStack <= 8 && code.Length < 64)
+			if (locals.Count == 0 && exceptions.Count == 0 && maxStack <= 8 && code.Length < 64 && !fatHeader)
 			{
 				rva = WriteTinyHeaderAndCode(bb);
 			}
