@@ -47,7 +47,7 @@ namespace IKVM.Reflection.Emit
 		private MethodImplAttributes implFlags;
 		private ILGenerator ilgen;
 		private int rva = -1;
-		private readonly CallingConventions callingConvention;
+		private CallingConventions callingConvention;
 		private List<ParameterBuilder> parameters;
 		private GenericTypeParameterBuilder[] gtpb;
 		private List<CustomAttributeBuilder> declarativeSecurity;
@@ -404,6 +404,11 @@ namespace IKVM.Reflection.Emit
 			this.attributes = attributes;
 		}
 
+		public void __SetCallingConvention(CallingConventions callingConvention)
+		{
+			this.callingConvention = callingConvention;
+		}
+
 		public override MethodImplAttributes GetMethodImplementationFlags()
 		{
 			return implFlags;
@@ -593,6 +598,11 @@ namespace IKVM.Reflection.Emit
 			set { initLocals = value; }
 		}
 
+		public void __AddUnmanagedExport(string name, int ordinal)
+		{
+			this.ModuleBuilder.AddUnmanagedExport(name, ordinal, this, new RelativeVirtualAddress(0xFFFFFFFF));
+		}
+
 		internal void Bake()
 		{
 			this.nameIndex = this.ModuleBuilder.Strings.Add(name);
