@@ -231,18 +231,12 @@ namespace IKVM.Reflection.Emit
 		private List<PropertyBuilder> properties;
 		private List<EventBuilder> events;
 		private TypeAttributes attribs;
-		private TypeFlags typeFlags;
 		private GenericTypeParameterBuilder[] gtpb;
 		private List<CustomAttributeBuilder> declarativeSecurity;
 		private List<Type> interfaces;
-
-		[Flags]
-		private enum TypeFlags
-		{
-			IsGenericTypeDefinition = 1,
-			HasNestedTypes = 2,
-			Baked = 4,
-		}
+		private int size;
+		private short pack;
+		private bool hasLayout;
 
 		internal TypeBuilder(ITypeOwner owner, string ns, string name)
 		{
@@ -252,6 +246,7 @@ namespace IKVM.Reflection.Emit
 			this.name = name;
 			this.typeNameSpace = ns == null ? 0 : this.ModuleBuilder.Strings.Add(ns);
 			this.typeName = this.ModuleBuilder.Strings.Add(name);
+			MarkEnumOrValueType(ns, name);
 		}
 
 		public ConstructorBuilder DefineDefaultConstructor(MethodAttributes attributes)
@@ -420,9 +415,12 @@ namespace IKVM.Reflection.Emit
 		public TypeBuilder DefineNestedType(string name, TypeAttributes attr, Type parent, Type[] interfaces)
 		{
 			TypeBuilder tb = DefineNestedType(name, attr, parent);
-			foreach (Type iface in interfaces)
+			if (interfaces != null)
 			{
-				tb.AddInterfaceImplementation(iface);
+				foreach (Type iface in interfaces)
+				{
+					tb.AddInterfaceImplementation(iface);
+				}
 			}
 			return tb;
 		}
@@ -454,7 +452,10 @@ namespace IKVM.Reflection.Emit
 			TypeBuilder typeBuilder = __DefineNestedType(ns, name);
 			typeBuilder.__SetAttributes(attr);
 			typeBuilder.SetParent(parent);
-			this.ModuleBuilder.SetPackingSizeAndTypeSize(typeBuilder, PackingSize.Unspecified, typeSize);
+			if (packSize != PackingSize.Unspecified || typeSize != 0)
+			{
+				typeBuilder.__SetLayout((int)packSize, typeSize);
+			}
 			return typeBuilder;
 		}
 
@@ -483,34 +484,33 @@ namespace IKVM.Reflection.Emit
 			interfaces.Add(interfaceType);
 		}
 
+		public void __SetInterfaceImplementationCustomAttribute(Type interfaceType, CustomAttributeBuilder cab)
+		{
+			this.ModuleBuilder.SetInterfaceImplementationCustomAttribute(this, interfaceType, cab);
+		}
+
 		public int Size
 		{
-			get
-			{
-				for (int i = 0; i < this.ModuleBuilder.ClassLayout.records.Length; i++)
-				{
-					if (this.ModuleBuilder.ClassLayout.records[i].Parent == token)
-					{
-						return this.ModuleBuilder.ClassLayout.records[i].ClassSize;
-					}
-				}
-				return 0;
-			}
+			get { return size; }
 		}
 
 		public PackingSize PackingSize
 		{
-			get
-			{
-				for (int i = 0; i < this.ModuleBuilder.ClassLayout.records.Length; i++)
-				{
-					if (this.ModuleBuilder.ClassLayout.records[i].Parent == token)
-					{
-						return (PackingSize)this.ModuleBuilder.ClassLayout.records[i].PackingSize;
-					}
-				}
-				return PackingSize.Unspecified;
-			}
+			get { return (PackingSize)pack; }
+		}
+
+		public override bool __GetLayout(out int packingSize, out int typeSize)
+		{
+			packingSize = this.pack;
+			typeSize = this.size;
+			return hasLayout;
+		}
+
+		public void __SetLayout(int packingSize, int typesize)
+		{
+			this.pack = (short)packingSize;
+			this.size = typesize;
+			this.hasLayout = true;
 		}
 
 		private void SetStructLayoutPseudoCustomAttribute(CustomAttributeBuilder customBuilder)
@@ -525,18 +525,12 @@ namespace IKVM.Reflection.Emit
 			{
 				layout = (LayoutKind)val;
 			}
-			int? pack = (int?)customBuilder.GetFieldValue("Pack");
-			int? size = (int?)customBuilder.GetFieldValue("Size");
-			if (pack.HasValue || size.HasValue)
-			{
-				ClassLayoutTable.Record rec = new ClassLayoutTable.Record();
-				rec.PackingSize = (short)(pack ?? 0);
-				rec.ClassSize = size ?? 0;
-				rec.Parent = token;
-				this.ModuleBuilder.ClassLayout.AddOrReplaceRecord(rec);
-			}
+			StructLayoutAttribute attr = new StructLayoutAttribute(layout);
+			attr.Pack = (int?)customBuilder.GetFieldValue("Pack") ?? 0;
+			attr.Size = (int?)customBuilder.GetFieldValue("Size") ?? 0;
+			attr.CharSet = customBuilder.GetFieldValue<CharSet>("CharSet") ?? CharSet.None;
 			attribs &= ~TypeAttributes.LayoutMask;
-			switch (layout)
+			switch (attr.Value)
 			{
 				case LayoutKind.Auto:
 					attribs |= TypeAttributes.AutoLayout;
@@ -548,9 +542,8 @@ namespace IKVM.Reflection.Emit
 					attribs |= TypeAttributes.SequentialLayout;
 					break;
 			}
-			CharSet? charSet = customBuilder.GetFieldValue<CharSet>("CharSet");
 			attribs &= ~TypeAttributes.StringFormatMask;
-			switch (charSet ?? CharSet.None)
+			switch (attr.CharSet)
 			{
 				case CharSet.None:
 				case CharSet.Ansi:
@@ -563,6 +556,9 @@ namespace IKVM.Reflection.Emit
 					attribs |= TypeAttributes.UnicodeClass;
 					break;
 			}
+			pack = (short)attr.Pack;
+			size = attr.Size;
+			hasLayout = pack != 0 || size != 0;
 		}
 
 		public void SetCustomAttribute(ConstructorInfo con, byte[] binaryAttribute)
@@ -665,6 +661,14 @@ namespace IKVM.Reflection.Emit
 				throw new NotImplementedException();
 			}
 			typeFlags |= TypeFlags.Baked;
+			if (hasLayout)
+			{
+				ClassLayoutTable.Record rec = new ClassLayoutTable.Record();
+				rec.PackingSize = pack;
+				rec.ClassSize = size;
+				rec.Parent = token;
+				this.ModuleBuilder.ClassLayout.AddRecord(rec);
+			}
 			foreach (MethodBuilder mb in methods)
 			{
 				mb.Bake();
@@ -818,20 +822,22 @@ namespace IKVM.Reflection.Emit
 		{
 			get
 			{
-				StructLayoutAttribute attr;
-				if ((attribs & TypeAttributes.ExplicitLayout) != 0)
+				LayoutKind layout;
+				switch (attribs & TypeAttributes.LayoutMask)
 				{
-					attr = new StructLayoutAttribute(LayoutKind.Explicit);
-					attr.Pack = 8;
-					attr.Size = 0;
-					this.ModuleBuilder.ClassLayout.GetLayout(token, ref attr.Pack, ref attr.Size);
-				}
-				else
-				{
-					attr = new StructLayoutAttribute((attribs & TypeAttributes.SequentialLayout) != 0 ? LayoutKind.Sequential : LayoutKind.Auto);
-					attr.Pack = 8;
-					attr.Size = 0;
+					case TypeAttributes.ExplicitLayout:
+						layout = LayoutKind.Explicit;
+						break;
+					case TypeAttributes.SequentialLayout:
+						layout = LayoutKind.Sequential;
+						break;
+					default:
+						layout = LayoutKind.Auto;
+						break;
 				}
+				StructLayoutAttribute attr = new StructLayoutAttribute(layout);
+				attr.Pack = (ushort)pack;
+				attr.Size = size;
 				switch (attribs & TypeAttributes.StringFormatMask)
 				{
 					case TypeAttributes.AutoClass:
@@ -843,6 +849,9 @@ namespace IKVM.Reflection.Emit
 					case TypeAttributes.AnsiClass:
 						attr.CharSet = CharSet.Ansi;
 						break;
+					default:
+						attr.CharSet = CharSet.None;
+						break;
 				}
 				return attr;
 			}
