@@ -125,6 +125,8 @@ namespace IKVM.Reflection
 		private Type typeof_System_Runtime_InteropServices_StructLayoutAttribute;
 		private Type typeof_System_Runtime_InteropServices_OptionalAttribute;
 		private Type typeof_System_Runtime_InteropServices_PreserveSigAttribute;
+		private Type typeof_System_Runtime_InteropServices_CallingConvention;
+		private Type typeof_System_Runtime_InteropServices_CharSet;
 		private Type typeof_System_Runtime_InteropServices_ComImportAttribute;
 		private Type typeof_System_Runtime_CompilerServices_DecimalConstantAttribute;
 		private Type typeof_System_Runtime_CompilerServices_SpecialNameAttribute;
@@ -151,6 +153,10 @@ namespace IKVM.Reflection
 
 		private Type ImportMscorlibType(System.Type type)
 		{
+			if (Mscorlib.__IsMissing)
+			{
+				return Mscorlib.ResolveType(new TypeName(type.Namespace, type.Name));
+			}
 			// We use FindType instead of ResolveType here, because on some versions of mscorlib some of
 			// the special types we use/support are missing and the type properties are defined to
 			// return null in that case.
@@ -161,7 +167,7 @@ namespace IKVM.Reflection
 		private Type ResolvePrimitive(string name)
 		{
 			// Primitive here means that these types have a special metadata encoding, which means that
-			// there can be references to them without refering to them by name explicitly.
+			// there can be references to them without referring to them by name explicitly.
 			// When 'resolve missing type' mode is enabled, we want these types to be usable even when
 			// they don't exist in mscorlib or there is no mscorlib loaded.
 			return Mscorlib.ResolveType(new TypeName("System", name));
@@ -367,6 +373,16 @@ namespace IKVM.Reflection
 			get { return typeof_System_Runtime_InteropServices_PreserveSigAttribute ?? (typeof_System_Runtime_InteropServices_PreserveSigAttribute = ImportMscorlibType(typeof(System.Runtime.InteropServices.PreserveSigAttribute))); }
 		}
 
+		internal Type System_Runtime_InteropServices_CallingConvention
+		{
+			get { return typeof_System_Runtime_InteropServices_CallingConvention ?? (typeof_System_Runtime_InteropServices_CallingConvention = ImportMscorlibType(typeof(System.Runtime.InteropServices.CallingConvention))); }
+		}
+
+		internal Type System_Runtime_InteropServices_CharSet
+		{
+			get { return typeof_System_Runtime_InteropServices_CharSet ?? (typeof_System_Runtime_InteropServices_CharSet = ImportMscorlibType(typeof(System.Runtime.InteropServices.CharSet))); }
+		}
+
 		internal Type System_Runtime_InteropServices_ComImportAttribute
 		{
 			get { return typeof_System_Runtime_InteropServices_ComImportAttribute ?? (typeof_System_Runtime_InteropServices_ComImportAttribute = ImportMscorlibType(typeof(System.Runtime.InteropServices.ComImportAttribute))); }
@@ -530,9 +546,20 @@ namespace IKVM.Reflection
 				}
 				return Import(type.GetGenericTypeDefinition()).MakeGenericType(importedArgs);
 			}
+			else if (type.IsNested)
+			{
+				// note that we can't pass in the namespace here, because .NET's Type.Namespace implementation is broken for nested types
+				// (it returns the namespace of the declaring type)
+				return Import(type.DeclaringType).ResolveNestedType(new TypeName(null, type.Name));
+			}
+			else if (type.Assembly == typeof(object).Assembly)
+			{
+				// make sure mscorlib types always end up in our mscorlib
+				return Mscorlib.ResolveType(new TypeName(type.Namespace, type.Name));
+			}
 			else
 			{
-				return Import(type.Assembly).GetType(type.FullName);
+				return Import(type.Assembly).ResolveType(new TypeName(type.Namespace, type.Name));
 			}
 		}
 
@@ -729,7 +756,7 @@ namespace IKVM.Reflection
 			{
 				return null;
 			}
-			return parser.GetType(this, context, throwOnError, assemblyQualifiedTypeName);
+			return parser.GetType(this, context, throwOnError, assemblyQualifiedTypeName, false);
 		}
 
 		public Assembly[] GetAssemblies()
@@ -811,7 +838,11 @@ namespace IKVM.Reflection
 		public Assembly CreateMissingAssembly(string assemblyName)
 		{
 			Assembly asm = new MissingAssembly(this, assemblyName);
-			assembliesByName.Add(asm.FullName, asm);
+			string name = asm.FullName;
+			if (!assembliesByName.ContainsKey(name))
+			{
+				assembliesByName.Add(name, asm);
+			}
 			return asm;
 		}
 
@@ -820,6 +851,11 @@ namespace IKVM.Reflection
 			resolveMissingMembers = true;
 		}
 
+		internal bool MissingMemberResolution
+		{
+			get { return resolveMissingMembers; }
+		}
+
 		private struct ScopedTypeName : IEquatable<ScopedTypeName>
 		{
 			private readonly object scope;
@@ -886,5 +922,25 @@ namespace IKVM.Reflection
 			}
 			throw new MissingMethodException(declaringType.ToString(), name);
 		}
+
+		internal FieldInfo GetMissingFieldOrThrow(Type declaringType, string name, FieldSignature signature)
+		{
+			if (resolveMissingMembers)
+			{
+				return new MissingField(declaringType, name, signature);
+			}
+			throw new MissingFieldException(declaringType.ToString(), name);
+		}
+
+		internal PropertyInfo GetMissingPropertyOrThrow(Type declaringType, string name, PropertySignature propertySignature)
+		{
+			// HACK we need to check __IsMissing here, because Type doesn't have a FindProperty API
+			// since properties are never resolved, except by custom attributes
+			if (resolveMissingMembers || declaringType.__IsMissing)
+			{
+				return new MissingProperty(declaringType, name, propertySignature);
+			}
+			throw new System.MissingMemberException(declaringType.ToString(), name);
+		}
 	}
 }
