@@ -78,6 +78,7 @@ namespace IKVM.Reflection.Writer
 			PEFileKinds fileKind, PortableExecutableKinds portableExecutableKind, ImageFileMachine imageFileMachine,
 			ResourceSection resources, int entryPointToken, Stream stream)
 		{
+			moduleBuilder.ApplyUnmanagedExports(imageFileMachine);
 			moduleBuilder.FixupMethodBodyTokens();
 
 			moduleBuilder.ModuleTable.Add(0, moduleBuilder.Strings.Add(moduleBuilder.moduleName), moduleBuilder.Guids.Add(moduleBuilder.ModuleVersionId), 0, 0);
@@ -99,13 +100,19 @@ namespace IKVM.Reflection.Writer
 				case ImageFileMachine.I386:
 					writer.Headers.FileHeader.Machine = IMAGE_FILE_HEADER.IMAGE_FILE_MACHINE_I386;
 					writer.Headers.FileHeader.Characteristics |= IMAGE_FILE_HEADER.IMAGE_FILE_32BIT_MACHINE;
+					writer.Headers.OptionalHeader.SizeOfStackReserve = moduleBuilder.GetStackReserve(0x100000);
+					break;
+				case ImageFileMachine.ARM:
+					writer.Headers.FileHeader.Machine = IMAGE_FILE_HEADER.IMAGE_FILE_MACHINE_ARM;
+					writer.Headers.FileHeader.Characteristics |= IMAGE_FILE_HEADER.IMAGE_FILE_32BIT_MACHINE;
+					writer.Headers.OptionalHeader.SizeOfStackReserve = moduleBuilder.GetStackReserve(0x100000);
 					break;
 				case ImageFileMachine.AMD64:
 					writer.Headers.FileHeader.Machine = IMAGE_FILE_HEADER.IMAGE_FILE_MACHINE_AMD64;
 					writer.Headers.FileHeader.Characteristics |= IMAGE_FILE_HEADER.IMAGE_FILE_LARGE_ADDRESS_AWARE;
 					writer.Headers.FileHeader.SizeOfOptionalHeader = 0xF0;
 					writer.Headers.OptionalHeader.Magic = IMAGE_OPTIONAL_HEADER.IMAGE_NT_OPTIONAL_HDR64_MAGIC;
-					writer.Headers.OptionalHeader.SizeOfStackReserve = 0x400000;
+					writer.Headers.OptionalHeader.SizeOfStackReserve = moduleBuilder.GetStackReserve(0x400000);
 					writer.Headers.OptionalHeader.SizeOfStackCommit = 0x4000;
 					writer.Headers.OptionalHeader.SizeOfHeapCommit = 0x2000;
 					break;
@@ -114,7 +121,7 @@ namespace IKVM.Reflection.Writer
 					writer.Headers.FileHeader.Characteristics |= IMAGE_FILE_HEADER.IMAGE_FILE_LARGE_ADDRESS_AWARE;
 					writer.Headers.FileHeader.SizeOfOptionalHeader = 0xF0;
 					writer.Headers.OptionalHeader.Magic = IMAGE_OPTIONAL_HEADER.IMAGE_NT_OPTIONAL_HDR64_MAGIC;
-					writer.Headers.OptionalHeader.SizeOfStackReserve = 0x400000;
+					writer.Headers.OptionalHeader.SizeOfStackReserve = moduleBuilder.GetStackReserve(0x400000);
 					writer.Headers.OptionalHeader.SizeOfStackCommit = 0x4000;
 					writer.Headers.OptionalHeader.SizeOfHeapCommit = 0x2000;
 					break;
@@ -153,6 +160,10 @@ namespace IKVM.Reflection.Writer
 			{
 				cliHeader.Flags |= CliHeader.COMIMAGE_FLAGS_32BITREQUIRED;
 			}
+			if ((portableExecutableKind & PortableExecutableKinds.Preferred32Bit) != 0)
+			{
+				cliHeader.Flags |= CliHeader.COMIMAGE_FLAGS_32BITREQUIRED | CliHeader.COMIMAGE_FLAGS_32BITPREFERRED;
+			}
 			if (keyPair != null)
 			{
 				cliHeader.Flags |= CliHeader.COMIMAGE_FLAGS_STRONGNAMESIGNED;
@@ -171,13 +182,26 @@ namespace IKVM.Reflection.Writer
 			moduleBuilder.Tables.Freeze(mw);
 			TextSection code = new TextSection(writer, cliHeader, moduleBuilder, ComputeStrongNameSignatureLength(publicKey));
 
+			// Export Directory
+			if (code.ExportDirectoryLength != 0)
+			{
+				writer.Headers.OptionalHeader.DataDirectory[0].VirtualAddress = code.ExportDirectoryRVA;
+				writer.Headers.OptionalHeader.DataDirectory[0].Size = code.ExportDirectoryLength;
+			}
+
 			// Import Directory
-			writer.Headers.OptionalHeader.DataDirectory[1].VirtualAddress = code.ImportDirectoryRVA;
-			writer.Headers.OptionalHeader.DataDirectory[1].Size = code.ImportDirectoryLength;
+			if (code.ImportDirectoryLength != 0)
+			{
+				writer.Headers.OptionalHeader.DataDirectory[1].VirtualAddress = code.ImportDirectoryRVA;
+				writer.Headers.OptionalHeader.DataDirectory[1].Size = code.ImportDirectoryLength;
+			}
 
 			// Import Address Table Directory
-			writer.Headers.OptionalHeader.DataDirectory[12].VirtualAddress = code.ImportAddressTableRVA;
-			writer.Headers.OptionalHeader.DataDirectory[12].Size = code.ImportAddressTableLength;
+			if (code.ImportAddressTableLength != 0)
+			{
+				writer.Headers.OptionalHeader.DataDirectory[12].VirtualAddress = code.ImportAddressTableRVA;
+				writer.Headers.OptionalHeader.DataDirectory[12].Size = code.ImportAddressTableLength;
+			}
 
 			// COM Descriptor Directory
 			writer.Headers.OptionalHeader.DataDirectory[14].VirtualAddress = code.ComDescriptorRVA;
@@ -190,15 +214,24 @@ namespace IKVM.Reflection.Writer
 				writer.Headers.OptionalHeader.DataDirectory[6].Size = code.DebugDirectoryLength;
 			}
 
-			writer.Headers.FileHeader.NumberOfSections = 2;
+			// we need to start by computing the number of sections, because code.PointerToRawData depends on that
+			writer.Headers.FileHeader.NumberOfSections = 1;
 
 			if (moduleBuilder.initializedData.Length != 0)
 			{
+				// .sdata
 				writer.Headers.FileHeader.NumberOfSections++;
 			}
 
-			if (resources != null && resources.Length != 0)
+			if (resources != null)
 			{
+				// .rsrc
+				writer.Headers.FileHeader.NumberOfSections++;
+			}
+
+			if (imageFileMachine != ImageFileMachine.ARM)
+			{
+				// .reloc
 				writer.Headers.FileHeader.NumberOfSections++;
 			}
 
@@ -236,14 +269,20 @@ namespace IKVM.Reflection.Writer
 			SectionHeader reloc = new SectionHeader();
 			reloc.Name = ".reloc";
 			reloc.VirtualAddress = rsrc.VirtualAddress + writer.ToSectionAlignment(rsrc.VirtualSize);
-			reloc.VirtualSize = 12;
+			if (imageFileMachine != ImageFileMachine.ARM)
+			{
+				reloc.VirtualSize = ((uint)moduleBuilder.unmanagedExports.Count + 1) * 12;
+			}
 			reloc.PointerToRawData = rsrc.PointerToRawData + rsrc.SizeOfRawData;
 			reloc.SizeOfRawData = writer.ToFileAlignment(reloc.VirtualSize);
 			reloc.Characteristics = SectionHeader.IMAGE_SCN_MEM_READ | SectionHeader.IMAGE_SCN_CNT_INITIALIZED_DATA | SectionHeader.IMAGE_SCN_MEM_DISCARDABLE;
 
-			// Base Relocation Directory
-			writer.Headers.OptionalHeader.DataDirectory[5].VirtualAddress = reloc.VirtualAddress;
-			writer.Headers.OptionalHeader.DataDirectory[5].Size = reloc.VirtualSize;
+			if (reloc.SizeOfRawData != 0)
+			{
+				// Base Relocation Directory
+				writer.Headers.OptionalHeader.DataDirectory[5].VirtualAddress = reloc.VirtualAddress;
+				writer.Headers.OptionalHeader.DataDirectory[5].Size = reloc.VirtualSize;
+			}
 
 			writer.Headers.OptionalHeader.SizeOfCode = text.SizeOfRawData;
 			writer.Headers.OptionalHeader.SizeOfInitializedData = sdata.SizeOfRawData + rsrc.SizeOfRawData + reloc.SizeOfRawData;
@@ -260,7 +299,7 @@ namespace IKVM.Reflection.Writer
 				// (i.e. there is an additional layer of indirection), so we add the offset to the pointer
 				writer.Headers.OptionalHeader.AddressOfEntryPoint = code.StartupStubRVA + 0x20;
 			}
-			else
+			else if (imageFileMachine != ImageFileMachine.ARM)
 			{
 				writer.Headers.OptionalHeader.AddressOfEntryPoint = code.StartupStubRVA;
 			}
@@ -275,13 +314,19 @@ namespace IKVM.Reflection.Writer
 			{
 				writer.WriteSectionHeader(rsrc);
 			}
-			writer.WriteSectionHeader(reloc);
+			if (reloc.SizeOfRawData != 0)
+			{
+				writer.WriteSectionHeader(reloc);
+			}
 
 			stream.Seek(text.PointerToRawData, SeekOrigin.Begin);
-			code.Write(mw, (int)sdata.VirtualAddress);
+			code.Write(mw, sdata.VirtualAddress);
 
-			stream.Seek(sdata.PointerToRawData, SeekOrigin.Begin);
-			mw.Write(moduleBuilder.initializedData);
+			if (sdata.SizeOfRawData != 0)
+			{
+				stream.Seek(sdata.PointerToRawData, SeekOrigin.Begin);
+				mw.Write(moduleBuilder.initializedData);
+			}
 
 			if (rsrc.SizeOfRawData != 0)
 			{
@@ -289,39 +334,14 @@ namespace IKVM.Reflection.Writer
 				resources.Write(mw, rsrc.VirtualAddress);
 			}
 
-			stream.Seek(reloc.PointerToRawData, SeekOrigin.Begin);
-			// .reloc section
-			uint relocAddress = code.StartupStubRVA;
-			switch (imageFileMachine)
+			if (reloc.SizeOfRawData != 0)
 			{
-				case ImageFileMachine.I386:
-				case ImageFileMachine.AMD64:
-					relocAddress += 2;
-					break;
-				case ImageFileMachine.IA64:
-					relocAddress += 0x20;
-					break;
-			}
-			uint pageRVA = relocAddress & ~0xFFFU;
-			mw.Write(pageRVA);	// PageRVA
-			mw.Write(0x000C);	// Block Size
-			if (imageFileMachine == ImageFileMachine.I386)
-			{
-				mw.Write(0x3000 + relocAddress - pageRVA);				// Type / Offset
-			}
-			else if (imageFileMachine == ImageFileMachine.AMD64)
-			{
-				mw.Write(0xA000 + relocAddress - pageRVA);				// Type / Offset
-			}
-			else if (imageFileMachine == ImageFileMachine.IA64)
-			{
-				// on IA64 the StartupStubRVA is 16 byte aligned, so these two addresses won't cross a page boundary
-				mw.Write((short)(0xA000 + relocAddress - pageRVA));		// Type / Offset
-				mw.Write((short)(0xA000 + relocAddress - pageRVA + 8));	// Type / Offset
+				stream.Seek(reloc.PointerToRawData, SeekOrigin.Begin);
+				code.WriteRelocations(mw);
 			}
 
 			// file alignment
-			mw.Write(new byte[writer.Headers.OptionalHeader.FileAlignment - reloc.VirtualSize]);
+			stream.SetLength(reloc.PointerToRawData + reloc.SizeOfRawData);
 
 			// do the strong naming
 			if (keyPair != null)
