@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2002-2008 Jeroen Frijters
+  Copyright (C) 2002-2011 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -828,6 +828,164 @@ namespace IKVM.Runtime
 		{
 			return ExceptionHelper.MapException<T>(x, (mode & MapFlags.NoRemapping) == 0, (mode & MapFlags.Unused) != 0);
 		}
+
+		public static T GetDelegateForInvokeExact<T>(global::java.lang.invoke.MethodHandle h)
+			where T : class
+		{
+#if FIRST_PASS
+			return null;
+#else
+			T del = h.vmtarget as T;
+			if (del == null)
+			{
+				throw new global::java.lang.invoke.WrongMethodTypeException();
+			}
+			return del;
+#endif
+		}
+
+		public static T GetDelegateForInvoke<T>(global::java.lang.invoke.MethodHandle h, ref InvokeCache<T> cache)
+			where T : class
+		{
+#if FIRST_PASS
+			return null;
+#else
+			if (cache.type == h.type() && cache.del != null)
+			{
+				return cache.del;
+			}
+			T del = h.vmtarget as T;
+			if (del == null)
+			{
+				global::java.lang.invoke.MethodHandle adapter = global::java.lang.invoke.MethodHandles.exactInvoker(h.type());
+				if (h.isVarargsCollector())
+				{
+					adapter = adapter.asVarargsCollector(h.type().parameterType(h.type().parameterCount() - 1));
+				}
+				del = (T)adapter.asType(MethodHandleUtil.GetDelegateMethodType(typeof(T))).vmtarget;
+				if (Interlocked.CompareExchange(ref cache.type, h.type(), null) == null)
+				{
+					cache.del = del;
+				}
+			}
+			return del;
+#endif
+		}
+
+		public static java.lang.invoke.MethodType LoadMethodType<T>()
+			where T : class // Delegate
+		{
+#if FIRST_PASS
+			return null;
+#else
+			return MethodHandleUtil.GetDelegateMethodType(typeof(T));
+#endif
+		}
+
+#if !FIRST_PASS
+		sealed class ConstantMethodHandle : java.lang.invoke.MethodHandle
+		{
+			internal ConstantMethodHandle(Delegate del)
+				: base(MethodHandleUtil.GetDelegateMethodType(del.GetType()))
+			{
+				this.vmtarget = del;
+			}
+		}
+#endif
+
+		public static java.lang.invoke.MethodHandle MethodHandleFromDelegate(Delegate del)
+		{
+#if FIRST_PASS
+			return null;
+#else
+			return new ConstantMethodHandle(del);
+#endif
+		}
+
+		[HideFromJava]
+		public static void LinkIndyCallSite<T>(ref IndyCallSite<T> site, java.lang.invoke.CallSite cs, Exception x)
+			where T : class // Delegate
+		{
+#if !FIRST_PASS
+			IndyCallSite<T> ics;
+			if (x != null || cs == null || (ics = cs.ics as IndyCallSite<T>) == null)
+			{
+				x = MapException<Exception>(x ?? (cs == null
+					? (Exception)new java.lang.ClassCastException("bootstrap method failed to produce a CallSite")
+					: new java.lang.invoke.WrongMethodTypeException()), MapFlags.None);
+				java.lang.invoke.MethodType type = LoadMethodType<T>();
+				ics = new IndyCallSite<T>((T)
+					java.lang.invoke.MethodHandles.dropArguments(
+						java.lang.invoke.MethodHandles.foldArguments(
+							java.lang.invoke.MethodHandles.throwException(type.returnType(), typeof(java.lang.BootstrapMethodError)),
+								new ConstantMethodHandle((MH<Exception, java.lang.BootstrapMethodError>)CreateBootstrapException).bindTo(x)),
+						0, type.parameterArray())
+					.vmtarget,
+					false);
+			}
+			IndyCallSite<T> curr = site;
+			if (curr.IsBootstrap)
+			{
+				Interlocked.CompareExchange(ref site, ics, curr);
+			}
+#endif
+		}
+
+#if !FIRST_PASS
+		[HideFromJava]
+		private static java.lang.BootstrapMethodError CreateBootstrapException(Exception x)
+		{
+			if (x is java.lang.BootstrapMethodError)
+			{
+				return (java.lang.BootstrapMethodError)x;
+			}
+			return new java.lang.BootstrapMethodError("call site initialization exception", x);
+		}
+#endif
+	}
+
+	public sealed class IndyCallSite<T>
+#if !FIRST_PASS
+		: java.lang.invoke.CallSite.IndyCallSite
+#endif
+		where T : class // Delegate
+	{
+		internal readonly bool IsBootstrap;
+		private volatile T target;
+
+		internal IndyCallSite()
+		{
+		}
+
+		internal IndyCallSite(T target, bool bootstrap)
+		{
+			this.IsBootstrap = bootstrap;
+			this.target = target;
+		}
+
+#if !FIRST_PASS
+		void java.lang.invoke.CallSite.IndyCallSite.setTarget(object target)
+		{
+			this.target = (T)target;
+		}
+#endif
+
+		public static IndyCallSite<T> CreateBootstrap(T bootstrap)
+		{
+			return new IndyCallSite<T>(bootstrap, true);
+		}
+
+		public T GetTarget()
+		{
+			return target;
+		}
+	}
+
+	public struct InvokeCache<T>
+		where T : class
+	{
+		internal global::java.lang.invoke.MethodType type;
+		internal T del;
 	}
 
 	[StructLayout(LayoutKind.Explicit)]
@@ -871,4 +1029,48 @@ namespace IKVM.Runtime
 			return converter.f;
 		}
 	}
+
+	public struct MHA<T1, T2, T3, T4, T5, T6, T7, T8>
+	{
+		public T1 t1;
+		public T2 t2;
+		public T3 t3;
+		public T4 t4;
+		public T5 t5;
+		public T6 t6;
+		public T7 t7;
+		public T8 t8;
+
+		public MHA(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5, T6 t6, T7 t7, T8 t8)
+		{
+			this.t1 = t1;
+			this.t2 = t2;
+			this.t3 = t3;
+			this.t4 = t4;
+			this.t5 = t5;
+			this.t6 = t6;
+			this.t7 = t7;
+			this.t8 = t8;
+		}
+	}
+
+	public delegate void MHV();
+	public delegate void MHV<T1>(T1 t1);
+	public delegate void MHV<T1, T2>(T1 t1, T2 t2);
+	public delegate void MHV<T1, T2, T3>(T1 t1, T2 t2, T3 t3);
+	public delegate void MHV<T1, T2, T3, T4>(T1 t1, T2 t2, T3 t3, T4 t4);
+	public delegate void MHV<T1, T2, T3, T4, T5>(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5);
+	public delegate void MHV<T1, T2, T3, T4, T5, T6>(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5, T6 t6);
+	public delegate void MHV<T1, T2, T3, T4, T5, T6, T7>(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5, T6 t6, T7 t7);
+	public delegate void MHV<T1, T2, T3, T4, T5, T6, T7, T8>(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5, T6 t6, T7 t7, T8 t8);
+
+	public delegate TResult MH<TResult>();
+	public delegate TResult MH<T1, TResult>(T1 t1);
+	public delegate TResult MH<T1, T2, TResult>(T1 t1, T2 t2);
+	public delegate TResult MH<T1, T2, T3, TResult>(T1 t1, T2 t2, T3 t3);
+	public delegate TResult MH<T1, T2, T3, T4, TResult>(T1 t1, T2 t2, T3 t3, T4 t4);
+	public delegate TResult MH<T1, T2, T3, T4, T5, TResult>(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5);
+	public delegate TResult MH<T1, T2, T3, T4, T5, T6, TResult>(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5, T6 t6);
+	public delegate TResult MH<T1, T2, T3, T4, T5, T6, T7, TResult>(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5, T6 t6, T7 t7);
+	public delegate TResult MH<T1, T2, T3, T4, T5, T6, T7, T8, TResult>(T1 t1, T2 t2, T3 t3, T4 t4, T5 t5, T6 t6, T7 t7, T8 t8);
 }
