@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2002-2010 Jeroen Frijters
+  Copyright (C) 2002-2011 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -37,7 +37,11 @@ namespace IKVM.Internal
 		NoSuchFieldError,
 		AbstractMethodError,
 		NoSuchMethodError,
-		LinkageError
+		LinkageError,
+		// "exceptions" that are wrapped in an IncompatibleClassChangeError
+		NoSuchFieldException,
+		NoSuchMethodException,
+		IllegalAccessException,
 		// if an error is added here, it must also be added to MethodAnalyzer.SetHardError()
 	}
 
@@ -72,11 +76,12 @@ namespace IKVM.Internal
 		private object[] annotations;
 		private string signature;
 		private string[] enclosingMethod;
+		private BootstrapMethod[] bootstrapMethods;
 
 		private static class SupportedVersions
 		{
 			internal static readonly int Minimum = 45;
-			internal static readonly int Maximum = JVM.SafeGetEnvironmentVariable("IKVM_EXPERIMENTAL_JDK_7") != null ? 51 : 50;
+			internal static readonly int Maximum = JVM.SafeGetEnvironmentVariable("IKVM_EXPERIMENTAL_JDK_8") != null ? 52 : 51;
 		}
 
 #if STATIC_COMPILER
@@ -119,12 +124,17 @@ namespace IKVM.Internal
 					case Constant.Fieldref:
 					case Constant.InterfaceMethodref:
 					case Constant.Methodref:
+					case Constant.InvokeDynamic:
 					case Constant.NameAndType:
 					case Constant.Float:
 					case Constant.Integer:
 						br.Skip(4);
 						break;
+					case Constant.MethodHandle:
+						br.Skip(3);
+						break;
 					case Constant.String:
+					case Constant.MethodType:
 						br.Skip(2);
 						break;
 					case Constant.Utf8:
@@ -203,6 +213,21 @@ namespace IKVM.Internal
 						case Constant.NameAndType:
 							constantpool[i] = new ConstantPoolItemNameAndType(br);
 							break;
+						case Constant.MethodHandle:
+							if (majorVersion < 51)
+								goto default;
+							constantpool[i] = new ConstantPoolItemMethodHandle(br);
+							break;
+						case Constant.MethodType:
+							if (majorVersion < 51)
+								goto default;
+							constantpool[i] = new ConstantPoolItemMethodType(br);
+							break;
+						case Constant.InvokeDynamic:
+							if (majorVersion < 51)
+								goto default;
+							constantpool[i] = new ConstantPoolItemInvokeDynamic(br);
+							break;
 						case Constant.String:
 							constantpool[i] = new ConstantPoolItemString(br);
 							break;
@@ -451,6 +476,13 @@ namespace IKVM.Internal
 							}
 							break;
 #endif
+						case "BootstrapMethods":
+							if(majorVersion < 51)
+							{
+								goto default;
+							}
+							bootstrapMethods = ReadBootstrapMethods(br, this);
+							break;
 						case "IKVM.NET.Assembly":
 							if(br.ReadUInt32() != 2)
 							{
@@ -463,6 +495,19 @@ namespace IKVM.Internal
 							break;
 					}
 				}
+				// validate the invokedynamic entries to point into the bootstrapMethods array
+				for(int i = 1; i < constantpoolcount; i++)
+				{
+					ConstantPoolItemInvokeDynamic cpi;
+					if(constantpool[i] != null
+						&& (cpi = constantpool[i] as ConstantPoolItemInvokeDynamic) != null)
+					{
+						if(bootstrapMethods == null || cpi.BootstrapMethod >= bootstrapMethods.Length)
+						{
+							throw new ClassFormatError("Short length on BootstrapMethods in class file");
+						}
+					}
+				}
 				// now that we've constructed the high level objects, the utf8 table isn't needed anymore
 				// TODO remove utf8_cp field from ClassFile object
 				utf8_cp = null;
@@ -490,6 +535,52 @@ namespace IKVM.Internal
 			//		}
 		}
 
+		private static BootstrapMethod[] ReadBootstrapMethods(BigEndianBinaryReader br, ClassFile classFile)
+		{
+			BigEndianBinaryReader rdr = br.Section(br.ReadUInt32());
+			ushort count = rdr.ReadUInt16();
+			BootstrapMethod[] bsm = new BootstrapMethod[count];
+			for(int i = 0; i < bsm.Length; i++)
+			{
+				ushort bsm_index = rdr.ReadUInt16();
+				if(bsm_index >= classFile.constantpool.Length || !(classFile.constantpool[bsm_index] is ConstantPoolItemMethodHandle))
+				{
+					throw new ClassFormatError("bootstrap_method_index {0} has bad constant type in class file {1}", bsm_index, classFile.Name);
+				}
+				ushort argument_count = rdr.ReadUInt16();
+				ushort[] args = new ushort[argument_count];
+				for(int j = 0; j < args.Length; j++)
+				{
+					ushort argument_index = rdr.ReadUInt16();
+					if(!classFile.IsValidConstant(argument_index))
+					{
+						throw new ClassFormatError("argument_index {0} has bad constant type in class file {1}", argument_index, classFile.Name);
+					}
+					args[j] = argument_index;
+				}
+				bsm[i] = new BootstrapMethod(bsm_index, args);
+			}
+			if(!rdr.IsAtEnd)
+			{
+				throw new ClassFormatError("Bad length on BootstrapMethods in class file {0}", classFile.Name);
+			}
+			return bsm;
+		}
+
+		private bool IsValidConstant(ushort index)
+		{
+			if(index < constantpool.Length && constantpool[index] != null)
+			{
+				try
+				{
+					constantpool[index].GetConstantType();
+					return true;
+				}
+				catch (InvalidOperationException) { }
+			}
+			return false;
+		}
+
 		private static object[] ReadAnnotations(BigEndianBinaryReader br, ClassFile classFile)
 		{
 			BigEndianBinaryReader rdr = br.Section(br.ReadUInt32());
@@ -892,6 +983,11 @@ namespace IKVM.Internal
 			return null;
 		}
 
+		internal ConstantPoolItemInvokeDynamic GetInvokeDynamic(int index)
+		{
+			return (ConstantPoolItemInvokeDynamic)constantpool[index];
+		}
+
 		private ConstantPoolItem GetConstantPoolItem(int index)
 		{
 			return constantpool[index];
@@ -902,9 +998,7 @@ namespace IKVM.Internal
 			return ((ConstantPoolItemClass)constantpool[index]).Name;
 		}
 
-		// this won't throw an exception if index is invalid
-		// (used by IsSideEffectFreeStaticInitializer)
-		internal bool SafeIsConstantPoolClass(int index)
+		private bool SafeIsConstantPoolClass(int index)
 		{
 			if(index > 0 && index < constantpool.Length)
 			{
@@ -965,6 +1059,16 @@ namespace IKVM.Internal
 			return ((ConstantPoolItemString)constantpool[index]).Value;
 		}
 
+		internal ConstantPoolItemMethodHandle GetConstantPoolConstantMethodHandle(int index)
+		{
+			return (ConstantPoolItemMethodHandle)constantpool[index];
+		}
+
+		internal ConstantPoolItemMethodType GetConstantPoolConstantMethodType(int index)
+		{
+			return (ConstantPoolItemMethodType)constantpool[index];
+		}
+
 		internal string Name
 		{
 			get
@@ -1096,6 +1200,38 @@ namespace IKVM.Internal
 			}
 		}
 
+		internal BootstrapMethod GetBootstrapMethod(int index)
+		{
+			return bootstrapMethods[index];
+		}
+
+		internal struct BootstrapMethod
+		{
+			private ushort bsm_index;
+			private ushort[] args;
+
+			internal BootstrapMethod(ushort bsm_index, ushort[] args)
+			{
+				this.bsm_index = bsm_index;
+				this.args = args;
+			}
+
+			internal int BootstrapMethodIndex
+			{
+				get { return bsm_index; }
+			}
+
+			internal int ArgumentCount
+			{
+				get { return args.Length; }
+			}
+
+			internal int GetArgument(int index)
+			{
+				return args[index];
+			}
+		}
+
 		internal struct InnerClass
 		{
 			internal ushort innerClass;		// ConstantPoolItemClass
@@ -1112,6 +1248,19 @@ namespace IKVM.Internal
 			}
 		}
 
+		internal enum RefKind
+		{
+			getField = 1,
+			getStatic = 2,
+			putField = 3,
+			putStatic = 4,
+			invokeVirtual = 5,
+			invokeStatic = 6,
+			invokeSpecial = 7,
+			newInvokeSpecial = 8,
+			invokeInterface = 9
+		}
+
 		internal enum ConstantType
 		{
 			Integer,
@@ -1119,7 +1268,9 @@ namespace IKVM.Internal
 			Float,
 			Double,
 			String,
-			Class
+			Class,
+			MethodHandle,
+			MethodType,
 		}
 
 		internal abstract class ConstantPoolItem
@@ -1335,6 +1486,8 @@ namespace IKVM.Internal
 			{
 				return clazz.GetClassType();
 			}
+
+			internal abstract MemberWrapper GetMember();
 		}
 
 		internal sealed class ConstantPoolItemFieldref : ConstantPoolItemFMI
@@ -1399,6 +1552,11 @@ namespace IKVM.Internal
 			{
 				return field;
 			}
+
+			internal override MemberWrapper GetMember()
+			{
+				return field;
+			}
 		}
 
 		internal class ConstantPoolItemMI : ConstantPoolItemFMI
@@ -1473,6 +1631,11 @@ namespace IKVM.Internal
 			{
 				return invokespecialMethod != null ? invokespecialMethod : method;
 			}
+
+			internal override MemberWrapper GetMember()
+			{
+				return method;
+			}
 		}
 
 		internal sealed class ConstantPoolItemMethodref : ConstantPoolItemMI
@@ -1641,6 +1804,224 @@ namespace IKVM.Internal
 			}
 		}
 
+		internal sealed class ConstantPoolItemMethodHandle : ConstantPoolItem
+		{
+			private byte ref_kind;
+			private ushort method_index;
+			private ConstantPoolItemFMI cpi;
+
+			internal ConstantPoolItemMethodHandle(BigEndianBinaryReader br)
+			{
+				ref_kind = br.ReadByte();
+				method_index = br.ReadUInt16();
+			}
+
+			internal override void Resolve(ClassFile classFile)
+			{
+				switch ((RefKind)ref_kind)
+				{
+					case RefKind.getField:
+					case RefKind.getStatic:
+					case RefKind.putField:
+					case RefKind.putStatic:
+						cpi = classFile.GetConstantPoolItem(method_index) as ConstantPoolItemFieldref;
+						break;
+					case RefKind.invokeSpecial:
+					case RefKind.invokeVirtual:
+					case RefKind.invokeStatic:
+					case RefKind.newInvokeSpecial:
+						cpi = classFile.GetConstantPoolItem(method_index) as ConstantPoolItemMethodref;
+						break;
+					case RefKind.invokeInterface:
+						cpi = classFile.GetConstantPoolItem(method_index) as ConstantPoolItemInterfaceMethodref;
+						break;
+				}
+				if (cpi == null)
+				{
+					throw new ClassFormatError("Invalid constant pool item MethodHandle");
+				}
+				if (ReferenceEquals(cpi.Name, StringConstants.INIT) && Kind != RefKind.newInvokeSpecial)
+				{
+					throw new ClassFormatError("Bad method name");
+				}
+			}
+
+			internal string Class
+			{
+				get { return cpi.Class; }
+			}
+
+			internal string Name
+			{
+				get { return cpi.Name; }
+			}
+
+			internal string Signature
+			{
+				get { return cpi.Signature; }
+			}
+
+			internal ConstantPoolItemFMI MemberConstantPoolItem
+			{
+				get { return cpi; }
+			}
+
+			internal RefKind Kind
+			{
+				get { return (RefKind)ref_kind; }
+			}
+
+			internal MemberWrapper Member
+			{
+				get { return cpi.GetMember(); }
+			}
+
+			internal TypeWrapper GetClassType()
+			{
+				return cpi.GetClassType();
+			}
+
+			internal override void Link(TypeWrapper thisType)
+			{
+				cpi.Link(thisType);
+			}
+
+			internal override ConstantType GetConstantType()
+			{
+				return ConstantType.MethodHandle;
+			}
+		}
+
+		internal sealed class ConstantPoolItemMethodType : ConstantPoolItem
+		{
+			private ushort signature_index;
+			private string descriptor;
+			private TypeWrapper[] argTypeWrappers;
+			private TypeWrapper retTypeWrapper;
+
+			internal ConstantPoolItemMethodType(BigEndianBinaryReader br)
+			{
+				signature_index = br.ReadUInt16();
+			}
+
+			internal override void Resolve(ClassFile classFile)
+			{
+				string descriptor = classFile.GetConstantPoolUtf8String(signature_index);
+				if (descriptor == null || !IsValidMethodSig(descriptor))
+				{
+					throw new ClassFormatError("Invalid MethodType signature");
+				}
+				this.descriptor = String.Intern(descriptor.Replace('/', '.'));
+			}
+
+			internal override void Link(TypeWrapper thisType)
+			{
+				lock (this)
+				{
+					if (argTypeWrappers != null)
+					{
+						return;
+					}
+				}
+				ClassLoaderWrapper classLoader = thisType.GetClassLoader();
+				TypeWrapper[] args = classLoader.ArgTypeWrapperListFromSigNoThrow(descriptor);
+				TypeWrapper ret = classLoader.RetTypeWrapperFromSigNoThrow(descriptor);
+				lock (this)
+				{
+					if (argTypeWrappers == null)
+					{
+						argTypeWrappers = args;
+						retTypeWrapper = ret;
+					}
+				}
+			}
+
+			internal TypeWrapper[] GetArgTypes()
+			{
+				return argTypeWrappers;
+			}
+
+			internal TypeWrapper GetRetType()
+			{
+				return retTypeWrapper;
+			}
+
+			internal override ConstantType GetConstantType()
+			{
+				return ConstantType.MethodType;
+			}
+		}
+
+		internal sealed class ConstantPoolItemInvokeDynamic : ConstantPoolItem
+		{
+			private ushort bootstrap_specifier_index;
+			private ushort name_and_type_index;
+			private string name;
+			private string descriptor;
+			private TypeWrapper[] argTypeWrappers;
+			private TypeWrapper retTypeWrapper;
+
+			internal ConstantPoolItemInvokeDynamic(BigEndianBinaryReader br)
+			{
+				bootstrap_specifier_index = br.ReadUInt16();
+				name_and_type_index = br.ReadUInt16();
+			}
+
+			internal override void Resolve(ClassFile classFile)
+			{
+				ConstantPoolItemNameAndType name_and_type = (ConstantPoolItemNameAndType)classFile.GetConstantPoolItem(name_and_type_index);
+				// if the constant pool items referred to were strings, GetConstantPoolItem returns null
+				if (name_and_type == null)
+				{
+					throw new ClassFormatError("Bad index in constant pool");
+				}
+				name = String.Intern(classFile.GetConstantPoolUtf8String(name_and_type.name_index));
+				descriptor = String.Intern(classFile.GetConstantPoolUtf8String(name_and_type.descriptor_index).Replace('/', '.'));
+			}
+
+			internal override void Link(TypeWrapper thisType)
+			{
+				lock (this)
+				{
+					if (argTypeWrappers != null)
+					{
+						return;
+					}
+				}
+				ClassLoaderWrapper classLoader = thisType.GetClassLoader();
+				TypeWrapper[] args = classLoader.ArgTypeWrapperListFromSigNoThrow(descriptor);
+				TypeWrapper ret = classLoader.RetTypeWrapperFromSigNoThrow(descriptor);
+				lock (this)
+				{
+					if (argTypeWrappers == null)
+					{
+						argTypeWrappers = args;
+						retTypeWrapper = ret;
+					}
+				}
+			}
+
+			internal TypeWrapper[] GetArgTypes()
+			{
+				return argTypeWrappers;
+			}
+
+			internal TypeWrapper GetRetType()
+			{
+				return retTypeWrapper;
+			}
+
+			internal string Name
+			{
+				get { return name; }
+			}
+
+			internal ushort BootstrapMethod
+			{
+				get { return bootstrap_specifier_index; }
+			}
+		}
+
 		private sealed class ConstantPoolItemString : ConstantPoolItem
 		{
 			private ushort string_index;
@@ -1682,7 +2063,10 @@ namespace IKVM.Internal
 			Fieldref = 9,
 			Methodref = 10,
 			InterfaceMethodref = 11,
-			NameAndType = 12
+			NameAndType = 12,
+			MethodHandle = 15,
+			MethodType = 16,
+			InvokeDynamic = 18,
 		}
 
 		internal abstract class FieldOrMethod
@@ -2070,7 +2454,8 @@ namespace IKVM.Internal
 			internal Method(ClassFile classFile, ClassFileParseOptions options, BigEndianBinaryReader br) : base(classFile, br)
 			{
 				// vmspec 4.6 says that all flags, except ACC_STRICT are ignored on <clinit>
-				if(ReferenceEquals(Name, StringConstants.CLINIT) && ReferenceEquals(Signature, StringConstants.SIG_VOID))
+				// however, since Java 7 it does need to be marked static
+				if(ReferenceEquals(Name, StringConstants.CLINIT) && ReferenceEquals(Signature, StringConstants.SIG_VOID) && (classFile.MajorVersion < 51 || IsStatic))
 				{
 					access_flags &= Modifiers.Strictfp;
 					access_flags |= (Modifiers.Static | Modifiers.Private);
@@ -2262,7 +2647,7 @@ namespace IKVM.Internal
 			{
 				get
 				{
-					return ReferenceEquals(Name, StringConstants.CLINIT) && ReferenceEquals(Signature, StringConstants.SIG_VOID);
+					return ReferenceEquals(Name, StringConstants.CLINIT) && ReferenceEquals(Signature, StringConstants.SIG_VOID) && IsStatic;
 				}
 			}
 
