@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2002-2010 Jeroen Frijters
+  Copyright (C) 2002-2011 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -47,6 +47,7 @@ namespace IKVM.Internal
 		private static readonly Dictionary<string, string> failedTypes = new Dictionary<string, string>();
 		private static readonly Key EXCEPTION_DATA_KEY = new Key();
 		private static readonly Exception NOT_REMAPPED = new Exception();
+		private static readonly Exception[] EMPTY_THROWABLE_ARRAY = new Exception[0];
 		private static readonly bool cleanStackTrace = JVM.SafeGetEnvironmentVariable("IKVM_DISABLE_STACKTRACE_CLEANING") == null;
 #if !FIRST_PASS
 		private static readonly ikvm.@internal.WeakIdentityMap exceptions = new ikvm.@internal.WeakIdentityMap();
@@ -326,20 +327,6 @@ namespace IKVM.Internal
 			return type.FullName;
 		}
 
-		private static void initThrowable(object throwable, object detailMessage, object cause)
-		{
-#if !FIRST_PASS
-			if(cause == throwable)
-			{
-				typeof(Throwable).GetConstructor(new Type[] { typeof(string) }).Invoke(throwable, new object[] { detailMessage });
-			}
-			else
-			{
-				typeof(Throwable).GetConstructor(new Type[] { typeof(string), typeof(Exception) }).Invoke(throwable, new object[] { detailMessage, cause });
-			}
-#endif
-		}
-
 		private static int GetLineNumber(StackFrame frame)
 		{
 			int ilOffset = frame.GetILOffset();
@@ -381,7 +368,8 @@ namespace IKVM.Internal
 			return new ObjectStreamField[] {
 				new ObjectStreamField("detailMessage", typeof(global::java.lang.String)),
 				new ObjectStreamField("cause", typeof(global::java.lang.Throwable)),
-				new ObjectStreamField("stackTrace", typeof(global::java.lang.StackTraceElement[]))
+				new ObjectStreamField("stackTrace", typeof(global::java.lang.StackTraceElement[])),
+				new ObjectStreamField("suppressedExceptions", typeof(global::java.util.List))
 			};
 #endif
 		}
@@ -392,14 +380,23 @@ namespace IKVM.Internal
 			lock (x)
 			{
 				ObjectOutputStream.PutField fields = s.putFields();
-				fields.put("detailMessage", Throwable.instancehelper_getMessage(x));
-				Exception cause = Throwable.instancehelper_getCause(x);
-				if (cause == null && x is Throwable)
+				Throwable _thisJava = x as Throwable;
+				if (_thisJava == null)
 				{
-					cause = ((Throwable)x).cause;
+					fields.put("detailMessage", x.Message);
+					fields.put("cause", x.InnerException);
+					// suppressed exceptions are not supported on CLR exceptions
+					fields.put("suppressedExceptions", null);
+					fields.put("stackTrace", getOurStackTrace(x));
+				}
+				else
+				{
+					fields.put("detailMessage", _thisJava.detailMessage);
+					fields.put("cause", _thisJava.cause);
+					fields.put("suppressedExceptions", _thisJava.suppressedExceptions);
+					getOurStackTrace(x);
+					fields.put("stackTrace", _thisJava.stackTrace ?? java.lang.ThrowableHelper.SentinelHolder.STACK_TRACE_SENTINEL);
 				}
-				fields.put("cause", cause);
-				fields.put("stackTrace", Throwable.instancehelper_getStackTrace(x));
 				s.writeFields();
 			}
 #endif
@@ -408,101 +405,86 @@ namespace IKVM.Internal
 		internal static void readObject(Exception x, ObjectInputStream s)
 		{
 #if !FIRST_PASS
-			ObjectInputStream.GetField fields = s.readFields();
-			initThrowable(x, fields.get("detailMessage", null), fields.get("cause", null));
-			StackTraceElement[] stackTrace = (StackTraceElement[])fields.get("stackTrace", null);
-			Throwable.instancehelper_setStackTrace(x, stackTrace == null ? new StackTraceElement[0] : stackTrace);
-#endif
-		}
-
-		internal static void printStackTrace(Exception x)
-		{
-#if !FIRST_PASS
-			Throwable.instancehelper_printStackTrace(x, java.lang.System.err);
-#endif
-		}
-
-		internal static void printStackTrace(Exception x, java.io.PrintStream printStream)
-		{
-#if !FIRST_PASS
-			lock (printStream)
+			lock (x)
 			{
-				foreach (string line in BuildStackTrace(x))
+				// when you serialize a .NET exception it gets replaced by a com.sun.xml.internal.ws.developer.ServerSideException,
+				// so we know that Exception is always a Throwable
+				Throwable _this = (Throwable)x;
+
+				// this the equivalent of s.defaultReadObject();
+				ObjectInputStream.GetField fields = s.readFields();
+				object detailMessage = fields.get("detailMessage", null);
+				object cause = fields.get("cause", null);
+				ConstructorInfo ctor = typeof(Throwable).GetConstructor(BindingFlags.NonPublic | BindingFlags.Instance, null, new Type[] { typeof(string), typeof(Exception), typeof(bool), typeof(bool) }, null);
+				if (cause == _this)
 				{
-					printStream.println(line);
+					ctor.Invoke(_this, new object[] { detailMessage, null, false, false });
+					_this.cause = _this;
 				}
-			}
-#endif
-		}
-
-		internal static void printStackTrace(Exception x, java.io.PrintWriter printWriter)
-		{
-#if !FIRST_PASS
-			lock (printWriter)
-			{
-				foreach (string line in BuildStackTrace(x))
+				else
 				{
-					printWriter.println(line);
+					ctor.Invoke(_this, new object[] { detailMessage, cause, false, false });
 				}
-			}
-#endif
-		}
+				_this.stackTrace = (StackTraceElement[])fields.get("stackTrace", null);
+				_this.suppressedExceptions = (java.util.List)fields.get("suppressedExceptions", null);
 
-#if !FIRST_PASS
-		private static List<String> BuildStackTrace(Exception x)
-		{
-			List<String> list = new List<String>();
-			list.Add(java.lang.Object.instancehelper_toString(x));
-			StackTraceElement[] stack = Throwable.instancehelper_getStackTrace(x);
-			for (int i = 0; i < stack.Length; i++)
-			{
-				list.Add("\tat " + stack[i]);
-			}
-			Exception cause = Throwable.instancehelper_getCause(x);
-			while (cause != null)
-			{
-				list.Add("Caused by: " + java.lang.Object.instancehelper_toString(cause));
-
-				// Cause stacktrace
-				StackTraceElement[] parentStack = stack;
-				stack = Throwable.instancehelper_getStackTrace(cause);
-				bool equal = false; // Is rest of stack equal to parent frame?
-				for (int i = 0; i < stack.Length && !equal; i++)
+				// this is where the rest of the Throwable.readObject() code starts
+				if (_this.suppressedExceptions != null)
 				{
-					// Check if we already printed the rest of the stack
-					// since it was the tail of the parent stack
-					int remaining = stack.Length - i;
-					int element = i;
-					int parentElement = parentStack.Length - remaining;
-					equal = parentElement >= 0 && parentElement < parentStack.Length;
-					while (equal && element < stack.Length)
+					java.util.List suppressed = null;
+					if (_this.suppressedExceptions.isEmpty())
 					{
-						if (stack[element].equals(parentStack[parentElement]))
-						{
-							element++;
-							parentElement++;
-						}
-						else
-						{
-							equal = false;
-						}
-					}
-					// Print stacktrace element or indicate the rest is equal 
-					if (!equal)
-					{
-						list.Add("\tat " + stack[i]);
+						suppressed = Throwable.SUPPRESSED_SENTINEL;
 					}
 					else
 					{
-						list.Add("\t... " + remaining + " more");
-						break; // from stack printing for loop
+						suppressed = new java.util.ArrayList(1);
+						for (int i = 0; i < _this.suppressedExceptions.size(); i++)
+						{
+							Exception entry = (Exception)_this.suppressedExceptions.get(i);
+							if (entry == null)
+							{
+								throw new java.lang.NullPointerException("Cannot suppress a null exception.");
+							}
+							if (entry == _this)
+							{
+								throw new java.lang.IllegalArgumentException("Self-suppression not permitted");
+							}
+							suppressed.add(entry);
+						}
+					}
+					_this.suppressedExceptions = suppressed;
+				}
+
+				if (_this.stackTrace != null)
+				{
+					if (_this.stackTrace.Length == 0)
+					{
+						_this.stackTrace = new StackTraceElement[0];
+					}
+					else if (_this.stackTrace.Length == 1
+						&& java.lang.ThrowableHelper.SentinelHolder.STACK_TRACE_ELEMENT_SENTINEL.equals(_this.stackTrace[0]))
+					{
+						_this.stackTrace = null;
+					}
+					else
+					{
+						foreach (StackTraceElement elem in _this.stackTrace)
+						{
+							if (elem == null)
+							{
+								throw new java.lang.NullPointerException("null StackTraceElement in serial stream. ");
+							}
+						}
 					}
 				}
-				cause = Throwable.instancehelper_getCause(cause);
+				else
+				{
+					_this.stackTrace = new StackTraceElement[0];
+				}
 			}
-			return list;
-		}
 #endif
+		}
 
 		internal static string FilterMessage(string message)
 		{
@@ -545,9 +527,17 @@ namespace IKVM.Internal
 #endif
 		}
 
-		internal static Exception getCause(Exception _this, Exception cause)
+		internal static Exception getCause(Exception _this)
 		{
-			return cause == _this ? null : cause;
+#if FIRST_PASS
+			return null;
+#else
+			lock (_this)
+			{
+				Exception cause = ((Throwable)_this).cause;
+				return cause == _this ? null : cause;
+			}
+#endif
 		}
 
 		internal static void checkInitCause(Exception _this, Exception _this_cause, Exception cause)
@@ -564,48 +554,124 @@ namespace IKVM.Internal
 #endif
 		}
 
-		internal static StackTraceElement[] computeStackTrace(Exception x, StackTrace part1, StackTrace part2)
+		internal static void addSuppressed(Exception _this, Exception x)
 		{
-#if FIRST_PASS
-			return null;
-#else
-			ExceptionInfoHelper eih = new ExceptionInfoHelper(part1, part2);
-			return eih.get_StackTrace(x);
-#endif
-		}
-
-		// this method is *only* for .NET exceptions (i.e. types not derived from java.lang.Throwable)
-		internal static StackTraceElement[] getStackTrace(Exception x)
-		{
-#if FIRST_PASS
-			return null;
-#else
-			lock (x)
+#if !FIRST_PASS
+			lock (_this)
 			{
-				ExceptionInfoHelper eih = null;
-				IDictionary data = x.Data;
-				if (data != null && !data.IsReadOnly)
+				if (_this == x)
 				{
-					lock (data.SyncRoot)
+					throw new java.lang.IllegalArgumentException("Self-suppression not permitted");
+				}
+				if (x == null)
+				{
+					throw new java.lang.NullPointerException("Cannot suppress a null exception.");
+				}
+				Throwable _thisJava = _this as Throwable;
+				if (_thisJava == null)
+				{
+					// we ignore suppressed exceptions for non-Java exceptions
+				}
+				else
+				{
+					if (_thisJava.suppressedExceptions == null)
 					{
-						eih = (ExceptionInfoHelper)data[EXCEPTION_DATA_KEY];
+						return;
 					}
+					if (_thisJava.suppressedExceptions == Throwable.SUPPRESSED_SENTINEL)
+					{
+						_thisJava.suppressedExceptions = new java.util.ArrayList();
+					}
+					_thisJava.suppressedExceptions.add(x);
 				}
-				if (eih == null)
-				{
-					return new StackTraceElement[0];
-				}
-				return eih.get_StackTrace(x);
 			}
 #endif
 		}
 
-		internal static StackTraceElement[] checkStackTrace(StackTraceElement[] original)
+		internal static Exception[] getSuppressed(Exception _this)
 		{
 #if FIRST_PASS
 			return null;
 #else
-			StackTraceElement[] copy = (StackTraceElement[])original.Clone();
+			lock (_this)
+			{
+				Throwable _thisJava = _this as Throwable;
+				if (_thisJava == null)
+				{
+					// we ignore suppressed exceptions for non-Java exceptions
+					return EMPTY_THROWABLE_ARRAY;
+				}
+				else
+				{
+					if (_thisJava.suppressedExceptions == Throwable.SUPPRESSED_SENTINEL
+						|| _thisJava.suppressedExceptions == null)
+					{
+						return EMPTY_THROWABLE_ARRAY;
+					}
+					return (Exception[])_thisJava.suppressedExceptions.toArray(EMPTY_THROWABLE_ARRAY);
+				}
+			}
+#endif
+		}
+
+		internal static int getStackTraceDepth(Exception _this)
+		{
+			return getOurStackTrace(_this).Length;
+		}
+
+		internal static StackTraceElement getStackTraceElement(Exception _this, int index)
+		{
+			return getOurStackTrace(_this)[index];
+		}
+
+		internal static StackTraceElement[] getOurStackTrace(Exception x)
+		{
+#if FIRST_PASS
+			return null;
+#else
+			Throwable _this = x as Throwable;
+			if (_this == null)
+			{
+				lock (x)
+				{
+					ExceptionInfoHelper eih = null;
+					IDictionary data = x.Data;
+					if (data != null && !data.IsReadOnly)
+					{
+						lock (data.SyncRoot)
+						{
+							eih = (ExceptionInfoHelper)data[EXCEPTION_DATA_KEY];
+						}
+					}
+					if (eih == null)
+					{
+						return Throwable.UNASSIGNED_STACK;
+					}
+					return eih.get_StackTrace(x);
+				}
+			}
+			else
+			{
+				lock (_this)
+				{
+					if (_this.stackTrace == Throwable.UNASSIGNED_STACK
+						|| (_this.stackTrace == null && (_this.tracePart1 != null || _this.tracePart2 != null)))
+					{
+						ExceptionInfoHelper eih = new ExceptionInfoHelper(_this.tracePart1, _this.tracePart2);
+						_this.stackTrace = eih.get_StackTrace(x);
+						_this.tracePart1 = null;
+						_this.tracePart2 = null;
+					}
+				}
+				return _this.stackTrace ?? Throwable.UNASSIGNED_STACK;
+			}
+#endif
+		}
+
+		internal static void setStackTrace(Exception x, StackTraceElement[] stackTrace)
+		{
+#if !FIRST_PASS
+			StackTraceElement[] copy = (StackTraceElement[])stackTrace.Clone();
 			for (int i = 0; i < copy.Length; i++)
 			{
 				if (copy[i] == null)
@@ -613,21 +679,35 @@ namespace IKVM.Internal
 					throw new java.lang.NullPointerException();
 				}
 			}
-			return copy;
+			SetStackTraceImpl(x, copy);
 #endif
 		}
 
-		// this method is *only* for .NET exceptions (i.e. types not derived from java.lang.Throwable)
-		internal static void setStackTrace(Exception x, StackTraceElement[] stackTrace)
+		private static void SetStackTraceImpl(Exception x, StackTraceElement[] stackTrace)
 		{
 #if !FIRST_PASS
-			ExceptionInfoHelper eih = new ExceptionInfoHelper(checkStackTrace(stackTrace));
-			IDictionary data = x.Data;
-			if (data != null && !data.IsReadOnly)
+			Throwable _this = x as Throwable;
+			if (_this == null)
 			{
-				lock (data.SyncRoot)
+				ExceptionInfoHelper eih = new ExceptionInfoHelper(stackTrace);
+				IDictionary data = x.Data;
+				if (data != null && !data.IsReadOnly)
 				{
-					data[EXCEPTION_DATA_KEY] = eih;
+					lock (data.SyncRoot)
+					{
+						data[EXCEPTION_DATA_KEY] = eih;
+					}
+				}
+			}
+			else
+			{
+				lock (_this)
+				{
+					if (_this.stackTrace == null && _this.tracePart1 == null && _this.tracePart2 == null)
+					{
+						return;
+					}
+					_this.stackTrace = stackTrace;
 				}
 			}
 #endif
@@ -767,7 +847,7 @@ namespace IKVM.Internal
 				Throwable t = x as Throwable;
 				if (t != null)
 				{
-					if (!unused && t.tracePart1 == null && t.tracePart2 == null && t.stackTrace == null)
+					if (!unused && t.tracePart1 == null && t.tracePart2 == null && t.stackTrace == Throwable.UNASSIGNED_STACK)
 					{
 						t.tracePart1 = new StackTrace(org, true);
 						t.tracePart2 = new StackTrace(true);
